import { mergeRight } from 'ramda';
import { middleware as googleMiddleware } from './analytics/google';
import { middleware as storyshareMiddleware } from './analytics/storyshare';
import { generateHexUUID } from './utils';

import {
  ANALYTICS_EVENT,
  ANALYTICS_INIT,
  ANALYTICS_PAGEVIEW,
  ANALYTICS_MAX_VISIT_DURATION,
  ANALYTICS_RANDOMISE_VISIT,
  ANALYTICS_SET,
  ANALYTICS_TYPE,
  PAUSE,
  RESUME,
  UNLOAD,
} from './const';

/**
 * Standardised Page View action creator
 * @param {string} resource - The resource as a path (eg /home)
 * @param {object} [args] - Argument object
 * @returns {AnyAction} - The generated action
 */
export const pageView = (resource, { ...props } = {}) => ({
  type: ANALYTICS_PAGEVIEW,
  resource,
  ...props,
});

/**
 * Standardised Event action creator
 * @param {number} type - The numeric type of the event (eg ANALYTICS_TYPE.SOCIAL)
 * @param {object} [args] - Argument object
 * @returns {AnyAction} - The generated action
 */
export const event = (type, { ...props } = {}) => ({
  type: ANALYTICS_EVENT,
  eventType: type,
  ...props,
});

export const init = () => ({
  type: ANALYTICS_INIT,
});

export const set = (payload) => ({
  type: ANALYTICS_SET,
  payload,
});

export const randomiseVisit = () => ({
  type: ANALYTICS_RANDOMISE_VISIT,
});

export const getVisitId = (state) => state.analytics.visitId;
export const getVisitStart = (state) => state.analytics.visitStart;

/**
 * This middleware handles adding generic types of data layer events to the analytics services
 * @param api
 */
export const middleware = ({ dispatch, getState }) => (next) => (action) => {
  switch (action.type) {
    case ANALYTICS_PAGEVIEW:
      // If the current visit has exceeded the maximum duration, create a new visit
      if ((getVisitStart(getState()) + ANALYTICS_MAX_VISIT_DURATION) < Date.now()) {
        dispatch(randomiseVisit());
      }
      break;

    case ANALYTICS_INIT:
      // Set up a new visit ID
      dispatch(randomiseVisit());
      break;

    case ANALYTICS_RANDOMISE_VISIT:
      dispatch(set({
        visitStart: Date.now(),
        visitId: generateHexUUID(),
      }));
      break;

    case PAUSE:
      dispatch(event(ANALYTICS_TYPE.APP, { action: 'paused' }));
      break;

    case RESUME:
      dispatch(event(ANALYTICS_TYPE.APP, { action: 'resumed' }));
      break;

    case UNLOAD:
      dispatch(event(ANALYTICS_TYPE.APP, { action: 'unload' }));
      break;
  }

  return next(action);
};

export const middlewares = [middleware, storyshareMiddleware, googleMiddleware];

export const reducer = (state = {}, action) => {
  switch (action.type) {
    case ANALYTICS_SET:
      return mergeRight(state, action.payload);
  }

  return state;
};
