import { path } from 'ramda';
import {
  ANALYTICS,
  ANALYTICS_EVENT,
  ANALYTICS_GOOGLE_PLATFORM as GOOGLE,
  ANALYTICS_INIT,
  ANALYTICS_PAGEVIEW,
  PLATFORM_CORDOVA,
  VERSION_APPCONTAINER,
} from '../const';
import { debug } from '../debug';
import { gaEvent, gaInit, gaPageView } from '../platforms/analytics/google';
import { platform } from '../utils';

/**
 * Google Analytics support
 * https://developers.google.com/analytics/devguides/collection/analyticsjs/field-reference
 */

export const middleware = ({ getState }) => (next) => (action) => {
  const state = getState();
  const site = path(['meta', 'site'], state);
  const gaId = path(['schema', 'keys', 'ga'], state);
  const appName = path(['schema', 'name'], state);
  const appVersion = path(['version', VERSION_APPCONTAINER], state);

  if (!site) {
    return next(action);
  }

  switch (action.type) {
    case ANALYTICS_INIT:
      gaInit({ gaId, appName, appVersion });
      break;

    case ANALYTICS_PAGEVIEW:
      debug(ANALYTICS, 'Google Analytics received pageview', action);

      // noinspection JSIgnoredPromiseFromCall
      gaPageView({ gaId, site, action });
      break;

    case ANALYTICS_EVENT:
      debug(ANALYTICS, 'Google Analytics received event', action);

      // noinspection JSIgnoredPromiseFromCall
      gaEvent({ site, action });
      break;
  }

  return next(action);
};

/**
 * Determine specific GA platform support type
 * @returns {boolean|number}
 */
export const googlePlatform = () => {
  if (typeof window === 'undefined') {
    return false;
  }

  switch (true) {
    case (platform() === PLATFORM_CORDOVA && !!window.ga):
      return GOOGLE.CORDOVA;

    case !!window.gtag:
      return GOOGLE.TAG_MANAGER;

    case !!window.ga:
      return GOOGLE.ANALYTICS;
  }

  return false;
};
