import { path } from 'ramda';
import { set, getVisitId } from '../analytics';
import {
  ANALYTICS,
  ANALYTICS_EVENT,
  ANALYTICS_PAGEVIEW,
  ANALYTICS_RANDOMISE_PAGEVIEW,
  SECOND,
  WEEK,
} from '../const';

import { debug } from '../debug';
import { getConnectionType } from '../platforms/utils';
import { generateHexUUID, getViewportSize, stripUndefinedProps } from '../utils';

export const randomisePageView = () => ({
  type: ANALYTICS_RANDOMISE_PAGEVIEW,
});

export const getPageViewId = (state) => state.analytics.pageViewId;

// This sentinel ensures we only send a referrer for the first pageView in this datalayer session
let sentReferrer = false;

/**
 * Sends a request to the service.
 * Uses a short TTL to remove the sent stat as it is no longer needed in the store once successfully deployed
 * Uses a try/catch in case the service that is receiving the stat does not support the analytics endpoints
 * @param site
 * @param type
 * @param params
 */
const sendRequest = (site, type, params = {}) => {
  try {
    site.req(`storyshare/v1/analytics/${type}`)
      .setParams(stripUndefinedProps(params))
      .retry(WEEK)
      .ttl(SECOND)
      .create();
  } catch (e) {
    // eslint-disable-next-line no-empty
  }
};

export const middleware = ({ dispatch, getState }) => (next) => (action) => {
  const site = path(['meta', 'site'], getState());
  let state;

  switch (action.type) {
    case ANALYTICS_PAGEVIEW:
      debug(ANALYTICS, 'Received pageview', action);

      // Create a new page view ID
      dispatch(randomisePageView());

      state = getState();

      const pvParams = {
        id: getPageViewId(state),
        timestamp: Date.now(),
        resource: action.resource,
        visit: getVisitId(state),
        title: action.title,
        connection_type: getConnectionType(),
      };

      const viewportSize = getViewportSize();
      if (viewportSize) {
        pvParams.resolution = `${viewportSize[ 'width' ]}x${viewportSize[ 'height' ]}`;
      }

      // Obtain the document referrer, but only for the first page view
      if (!sentReferrer && typeof document === 'object' && document.referrer) {
        pvParams.referrer = document.referrer;
      }
      sentReferrer = true;

      // Attempt to transform a short-form object route such as article/4 into a complete REST endpoint
      // This may throw if an invalid objectRoute was provided
      try {
        pvParams.object_route = site.req(action.objectRoute).self;
      } catch (e) {
        // eslint-disable-next-line no-empty
      }

      sendRequest(site, 'pageview', pvParams);
      break;

    case ANALYTICS_EVENT:
      debug(ANALYTICS, 'Received event', action);

      state = getState();

      const eParams = {
        id: generateHexUUID(),
        pageview: getPageViewId(state),
        timestamp: Date.now(),
        type: action.eventType,
        action: action.action,
        value: action.value,
        label: action.label,
        connection_type: getConnectionType(),
      };

      try {
        eParams.object_route = site.req(action.objectRoute).self;
      } catch (e) {
        // eslint-disable-next-line no-empty
      }

      sendRequest(site, 'event', eParams);
      break;

    case ANALYTICS_RANDOMISE_PAGEVIEW:
      dispatch(set({
        pageViewId: generateHexUUID(),
      }));
      break;
  }

  return next(action);
};
