import { assoc, dissoc, forEach, keys, path } from 'ramda';
import URI from 'urijs';

import { ASSET, ASSET_ADD, ASSET_DROP, QUEUE_PROCESS } from './const';
import { error } from './debug';
import * as platform from './platforms/asset';
import { getExpiredAssets } from './selector';
import { deleteFile } from './storage';
import { externalFlag, stripQuery } from './utils';

export const add = (href, payload) => ({
  type: ASSET_ADD,
  href,
  payload,
});

export const drop = (href) => ({
  type: ASSET_DROP,
  href,
});

export const middleware = ({ dispatch, getState }) => (next) => (action) => {
  switch (action.type) {
    case ASSET_DROP:
      const filename = path(['assets', action.href, 'filename'], getState());
      if (filename) {
        deleteFile(filename).catch((e) => {
          error(ASSET, 'Received error deleting file', filename, e);
        });
      }
      break;

    case QUEUE_PROCESS:
      const expiredUrls = keys(getExpiredAssets(getState()));
      forEach((href) => dispatch(drop(href)), expiredUrls);
      break;
  }

  return next(action);
};

export const reducer = (state = {}, action) => {
  switch (action.type) {
    case ASSET_ADD:
      return assoc(
        // remove query components to produce a canonical asset URL
        stripQuery(action.href),
        // tag the payload with the current time so we can detect expiry
        assoc('now', Date.now(), action.payload),
        state,
      );

    case ASSET_DROP:
      return dissoc(action.href, state);
  }

  return state;
};

export const download = ({ url }) => {
  if (externalFlag('simulateOffline')) {
    throw new TypeError(`Offline simulation failing request for: ${URI(url).toString()}`);
  }

  return platform.download(url);
};
