import { defaultTo, merge, omit, path, pick, prop } from 'ramda';
import { event as trackEvent } from './analytics';

import {
  ANALYTICS_TYPE,
  AUTH,
  AUTH_JWT_TOKEN_EXPIRE_ENDPOINT,
  AUTH_JWT_TOKEN_GET_ENDPOINT,
  AUTH_LOGGED_IN,
  AUTH_LOGGED_OUT,
  AUTH_LOGIN_AUTOMATIC,
  AUTH_LOGIN_CREDENTIALS,
  AUTH_LOGIN_JWT,
  AUTH_SET,
  AUTH_STATUS_LOGGED_IN,
  AUTH_STATUS_LOGGED_OUT,
  AUTH_STATUS_PENDING,
  AUTH_STATUS_UNKNOWN,
  ERROR_USAGE,
  EVENT_AUTH_IDENTITY_CHANGED,
  EVENT_AUTH_LOGGED_IN,
  EVENT_AUTH_LOGGED_OUT,
} from './const';
import { debug } from './debug';
import Error from './error';
import { generateEvent } from './event';
import * as platform from './platforms/auth';
import { Promise } from './promise';
import { fetch as fetchSchema } from './schema';
import { base64encode } from './utils';

export const set = (meta) => ({
  type: AUTH_SET,
  meta,
});

/**
 * Action creator to set the auth status
 * @param {AuthStatus} status - authentication status
 * @returns {AnyAction}
 */
export const setStatus = (status) => set({ status });

/**
 * Selector to read the auth status
 * @param {Store} state - store state
 * @returns {AuthStatus} authentication status
 */
export const getStatus = (state) => state.auth.status || AUTH_STATUS_UNKNOWN;

/**
 * Action creator to store credentials for HTTP Basic authentication
 * @param {string} username - HTTP Basic username
 * @param {string} password - HTTP Basic password
 * @returns {AnyAction}
 */
export const setBasicCredentials = (username, password) =>
  set({ basic: base64encode(`${username}:${password}`) });

export const loggedIn = (user) =>
  ({ type: AUTH_LOGGED_IN, user });

export const loggedOut = () =>
  ({ type: AUTH_LOGGED_OUT });

/**
 * Action creator to store credentials for JWT authentication
 * @param {string} token - JWT token
 * @returns {AnyAction}
 */
export const setTokenCredentials = (token) =>
  set({ jwt: token });

/**
 * Action creator to attempt authentication
 * @param {Credentials} credentials - credentials to try
 * @returns {AnyAction}
 */
export const authenticate = ({ jwt, email, username, password } = {}) => {
  if (jwt) {
    return {
      type: AUTH_LOGIN_JWT,
      jwt,
    };
  }

  if (email) {
    return {
      type: AUTH_LOGIN_CREDENTIALS,
      email,
      password,
    };
  }

  if (username) {
    return {
      type: AUTH_LOGIN_CREDENTIALS,
      username,
      password,
    };
  }

  return {
    type: AUTH_LOGIN_AUTOMATIC,
  };
};

export const middleware = ({ getState, dispatch }) => (next) => (action) => {
  const store = getState();
  const { site } = store.meta;

  switch (action.type) {
    case AUTH_LOGGED_IN:
      debug(AUTH, 'Logged in');

      const currentUserId = prop('id', defaultTo(null, path(['auth', 'user'], store)));

      if (getStatus(store) === AUTH_STATUS_UNKNOWN) {
        break;
      }

      // If the user's identity has not changed
      if (currentUserId && currentUserId === path(['user', 'id'], action)) {
        break;
      }

      // If the user's identity has changed
      dispatch(generateEvent(EVENT_AUTH_IDENTITY_CHANGED));
      dispatch(generateEvent(EVENT_AUTH_LOGGED_IN, action.user));
      dispatch(trackEvent(ANALYTICS_TYPE.AUTH, { action: 'login', interactive: false }));
      break;

    case AUTH_LOGGED_OUT:
      debug(AUTH, 'Logged out');

      if (store.auth.jwt) {
        site.req(AUTH_JWT_TOKEN_EXPIRE_ENDPOINT)
          .addParam('token', store.auth.jwt)
          .create();
      }

      dispatch(generateEvent(EVENT_AUTH_IDENTITY_CHANGED));
      dispatch(generateEvent(EVENT_AUTH_LOGGED_OUT));
      dispatch(trackEvent(ANALYTICS_TYPE.AUTH, { action: 'logout', interactive: false }));
      break;

    case AUTH_LOGIN_JWT:
      debug(AUTH, 'Token login');

      if (!action.jwt) {
        throw new Error(ERROR_USAGE, 'No JSON Web Token was provided');
      }

      dispatch(setTokenCredentials(action.jwt));

      return dispatch(fetchSchema())
        .then(() => site.isLoggedIn(false));

    case AUTH_LOGIN_CREDENTIALS:
      debug(AUTH, 'Credential login');

      dispatch(setStatus(AUTH_STATUS_PENDING));

      return site.req(AUTH_JWT_TOKEN_GET_ENDPOINT)
        .setParams(pick(['email', 'username', 'password'], action))
        .promise()
        .create()
        .then((response) => {
          dispatch(setTokenCredentials(response.token));

          return dispatch(fetchSchema())
            .then(() => site.isLoggedIn(false));
        });

    case AUTH_LOGIN_AUTOMATIC:
      if (getStatus(getState()) === AUTH_STATUS_PENDING) {
        return Promise.resolve();
      }

      debug(AUTH, 'Automatic login');

      return platform.loginAutomatic({ dispatch, getState })
        .then(() => dispatch(fetchSchema()))
        .then(() => site.isLoggedIn(false));
  }

  return next(action);
};

export const reducer = (state = {}, action) => {
  switch (action.type) {
    case AUTH_SET:
      return merge(state, action.meta);

    case AUTH_LOGGED_IN:
      return merge(state, {
        user: action.user,
        status: AUTH_STATUS_LOGGED_IN,
      });

    case AUTH_LOGGED_OUT:
      return merge(
        omit(['jwt', 'basic', 'user'], state),
        {
          status: AUTH_STATUS_LOGGED_OUT,
        },
      );
  }

  return state;
};
