import { merge } from 'ramda';

import {
  AUTOUPDATE,
  AUTOUPDATE_CHECK,
  AUTOUPDATE_INIT,
  AUTOUPDATE_LOAD_ASSETS,
  AUTOUPDATE_REDIRECT,
  AUTOUPDATE_SET,
  AUTOUPDATE_STATUS_DISABLED,
  AUTOUPDATE_STATUS_IDLE,
  AUTOUPDATE_STATUS_PENDING_RESTART,
  AUTOUPDATE_STATUS_UPDATING,
  RESUME,
  SOFT_AUTOUPDATE_FREQ,
} from './const';
import { debug, warn } from './debug';
import * as platform from './platforms/autoupdate';
import { Promise } from './promise';
import { externalFlag } from './utils';

export const init = (repository) => ({
  type: AUTOUPDATE_INIT,
  repository,
});

export const check = (force) => ({
  type: AUTOUPDATE_CHECK,
  force,
});

export const redirect = () => ({
  type: AUTOUPDATE_REDIRECT,
});

export const set = (meta) => ({
  type: AUTOUPDATE_SET,
  meta,
});

export const loadAssets = () => ({
  type: AUTOUPDATE_LOAD_ASSETS,
});

export const storeHashes = (hashes) => set({ hashes });
export const setStatus = (status) => set({ status });
export const setRepository = (repository) => set({ repository });
export const setProgress = (progress) => set({ progress });
export const setLastChecked = () => set({ lastChecked: Date.now() });
export const setLoad = (load) => set({ load });
export const setActive = (active) => set({ active });

export const getProgress = (state) => state.autoupdate.progress;
export const getStatus = (state) => state.autoupdate.status;
export const getHashes = (state) => state.autoupdate.hashes;
export const getLastChecked = (state) => state.autoupdate.lastChecked;
export const getRepository = (state) => state.autoupdate.repository;
export const getActive = (state) => state.autoupdate.active;
export const getLoad = (state) => state.autoupdate.load;

export const middleware = ({ dispatch, getState }) => (next) => (action) => {
  switch (action.type) {
    case AUTOUPDATE_INIT:
      return (async () => {
        // If autoupdate is disabled then just return
        if (externalFlag('disableAutoupdate')) {
          dispatch(setStatus(AUTOUPDATE_STATUS_DISABLED));
          return;
        }

        // Set the initial state of autoupdate
        dispatch(setStatus(AUTOUPDATE_STATUS_IDLE));
        dispatch(setRepository(action.repository));

        // Perform platform-specific initialisation
        await platform.init({ dispatch, getState });

        // If autoupdate has never run then await its completion
        if (getLastChecked(getState()) === null) {
          await dispatch(check(true));
        }

        // Post-init, if a reboot is pending do that now
        if (getStatus(getState()) === AUTOUPDATE_STATUS_PENDING_RESTART) {
          await dispatch(redirect());
        }

        // Load any platform-specific assets
        dispatch(loadAssets());
      })();

    case AUTOUPDATE_REDIRECT:
      if (getStatus(getState()) !== AUTOUPDATE_STATUS_PENDING_RESTART) {
        return Promise.resolve(false);
      }

      return platform.redirect({ dispatch, getState });

    case RESUME:
      if (getStatus(getState()) === AUTOUPDATE_STATUS_PENDING_RESTART) {
        return dispatch(redirect());
      }

      dispatch(check());
      break;

    case AUTOUPDATE_LOAD_ASSETS:
      platform.loadAssets({ getState });
      break;

    case AUTOUPDATE_CHECK:
      if (getStatus(getState()) === AUTOUPDATE_STATUS_DISABLED) {
        break;
      }

      const repository = getRepository(getState());
      if (!repository) {
        warn(AUTOUPDATE, 'Skipping autoupdate, no repository is defined');
        break;
      }

      if (getStatus(getState()) === AUTOUPDATE_STATUS_UPDATING) {
        debug(AUTOUPDATE, 'Autoupdate was already running');
        break;
      }

      if (getStatus(getState()) === AUTOUPDATE_STATUS_PENDING_RESTART) {
        debug(AUTOUPDATE, 'Autoupdate was already pending restart');
        break;
      }

      const lastChecked = getLastChecked(getState());
      if (!action.force && (lastChecked + SOFT_AUTOUPDATE_FREQ > Date.now())) {
        break;
      }

      return platform.check({ dispatch, getState });
  }

  return next(action);
};

export const reducer = (state = {
  status: AUTOUPDATE_STATUS_IDLE,
  progress: 0,
  lastChecked: null,
  hashes: {},
  active: null,
  load: [],
}, action) => {
  switch (action.type) {
    case AUTOUPDATE_SET:
      return merge(state, action.meta);
  }

  return state;
};
