import 'cross-fetch/polyfill';

import {
  append,
  assocPath,
  concat,
  contains,
  defaultTo,
  drop,
  endsWith,
  fromPairs,
  groupBy,
  has,
  head,
  isEmpty,
  keys,
  length,
  map,
  merge,
  omit,
  path,
  pathEq,
  pick,
  pipe,
  prop,
  propEq,
  propOr,
  reduce,
  splitEvery,
  takeWhile,
  toLower,
  toPairs,
  values,
  zipObj,
  zipWith,
} from 'ramda';
import URI from 'urijs';

import { add as addAsset, download } from './asset';
import { loggedIn, loggedOut } from './auth';
import {
  AUTH_ME_ENDPOINT,
  ERROR_INTERNAL,
  FETCH,
  FETCH_SET,
  FETCH_SET_HEADER,
  HTTP_CREATED,
  HTTP_OK,
  HTTP_UNAUTHORIZED,
  METHOD_CREATE,
  METHOD_DELETE,
  METHOD_DOWNLOAD,
  METHOD_MULTI,
  METHOD_READ,
  METHOD_UPDATE,
  METHOD_UPLOAD,
  QUEUE_PROCESS,
  REQUESTS_PER_MULTI,
  STATUS_BACKOFF,
  STATUS_FAILED,
  STATUS_INFLIGHT,
  STATUS_READY,
  STATUS_RETRY,
  STATUS_SUCCESS,
} from './const';
import { warn } from './debug';
import Error from './error';
import { add as addObject, drop as dropObject } from './object';
import * as platform from './platforms/fetch';
import { Promise, reject, resolve } from './promise';
import { dequeueRequest, setRequestStatus, updateRequest } from './queue';
import { getObjectBySelf, getRequestsByStatus } from './selector';
import { backoff, externalFlag, httpMethod, maybeMap } from './utils';

export const set = (meta) => ({
  type: FETCH_SET,
  meta,
});

/**
 * Action creator to set a default header value
 * @param {string} name - header name
 * @param {string} value - header value
 * @returns {AnyAction}
 */
export const setHeader = (name, value) => ({
  type: FETCH_SET_HEADER,
  name,
  value,
});

/**
 * Selector to retrieve all set headers
 * @param {Store} state - store state
 * @returns { { [index: string]: string } }
 */
export const getHeaders = (state) => state.fetch.headers;

/**
 * Selector to retrieve credentials for use making requests
 * @param {Store} state - store state
 * @returns { { basic: any, jwt: any } }
 */
export const getCredential = (state) => pick(['basic', 'jwt'], state.auth);

export const fetchWrapper = async (
  {
    url,
    method = METHOD_READ,
    body = {},
    query = {},
    headers = {},
    auth = {},
  },
) => {
  const uri = new URI(url);

  if (auth.jwt) {
    headers[ 'Authorization' ] = `Bearer ${auth.jwt}`;
  }

  if (auth.basic) {
    headers[ 'Authorization' ] = `Basic ${auth.basic}`;
  }

  const options = {
    method: httpMethod(method),
    credentials: 'include',
    headers,
  };

  switch (method) {
    case METHOD_UPLOAD:
      options.headers[ 'Accept' ] = 'application/json';
      query = merge(query, omit(['file'], body));
      await platform.prepareForUpload(body, options);
      break;

    case METHOD_CREATE:
    case METHOD_UPDATE:
    case METHOD_MULTI:
      options.body = JSON.stringify(body);
      options.headers[ 'Content-Type' ] = 'application/json';
      options.headers[ 'Accept' ] = 'application/json';
      break;
  }

  if (!isEmpty(query)) {
    uri.setSearch(query);
  }

  if (externalFlag('simulateOffline')) {
    throw new TypeError(`Offline simulation failing request for: ${uri.toString()}`);
  }

  const response = await fetch(uri.toString(), options);

  return {
    response,
    result: await response.json(),
  };
};

export const middleware = ({ getState, dispatch }) => (next) => (action) => {
  if (action.type !== QUEUE_PROCESS) {
    return next(action);
  }

  const { site, endpoint } = pick(['site', 'endpoint'], prop('meta', getState()));

  // Move from retry to backoff to ready
  map(async (request) => {
    // Check if the request has been retried for too long
    if ((request.retry + request.now) < Date.now()) {
      dispatch(setRequestStatus(request.hash, STATUS_FAILED));
      dispatch(reject(request.hash, `Request for ${request.url} could not be completed during the retry interval`));
      return;
    }

    if (request.retry === false) {
      dispatch(setRequestStatus(request.hash, STATUS_FAILED));
      dispatch(reject(request.hash, `Request for ${request.url} failed and was not configured to retry`));
      return;
    }

    let attempt = defaultTo(1, request.attempt);
    dispatch(setRequestStatus(request.hash, STATUS_BACKOFF));
    await Promise.delay(backoff(++attempt));
    dispatch(updateRequest(request.hash, { attempt, status: STATUS_RETRY }));
  }, getRequestsByStatus(STATUS_RETRY, getState()));

  // Gets the ready queue
  let readyQueue = getRequestsByStatus(STATUS_READY, getState());

  if (isEmpty(readyQueue)) {
    return next(action);
  }

  // Ensure query parameter keys are properly encoded using [] syntax for arrays
  readyQueue = map((request) => merge(request, {
    query: pipe(toPairs, map((pair) =>
      [Array.isArray(pair[ 1 ]) && !endsWith('[]', pair[ 0 ]) ? `${pair[ 0 ]}[]` : pair[ 0 ], pair[ 1 ]],
    ), fromPairs)(request.query),
  }), readyQueue);

  // Selects a block of requests that have the same method
  const method = prop('method', head(readyQueue));

  // Test whether this service supports multi
  let hasMulti = has('/wp/v2/multi', defaultTo({}, site._routes));

  // Optionally disable multi
  if (defaultTo(true, site._options.useMulti) === false) {
    hasMulti = false;
  }

  // Splits the requests into ones that can be bulk processed, and ones that cannot
  // noinspection JSUnresolvedFunction
  let { multi = [], std = [] } = groupBy(
    (request) => (
      hasMulti
      && request.method !== METHOD_UPLOAD // File uploads cannot use multi
      && contains('wp-json/', URI(request.url).path()) // wp-json endpoints can be sent using multi
    ) ? 'multi' : 'std',
    takeWhile((request) => propEq('method', method, request), readyQueue),
  );

  if (length(multi) < 2) {
    std = concat(multi, std);
    multi = [];
  }

  // Map the requests into runnable blocks of:
  // [{ path: '/some/path', headers: [], body: {}, method: METHOD_X },{}...]

  // noinspection JSUnresolvedFunction
  let requests = map((block) => ({
    requests: block,
    fetchable: {
      url: `${endpoint}/wp/v2/multi`,
      method: METHOD_MULTI,
      body: {
        requests: reduce((acc, request) =>
          append({
            path: URI(drop(length(endpoint), request.url)).query(defaultTo({}, request.query)).toString(),
            body: defaultTo({}, request.body),
            method: httpMethod(request.method),
            headers: defaultTo({}, request.headers),
          }, acc), [], block),
      },
    },
  }), splitEvery(REQUESTS_PER_MULTI, multi));

  requests = reduce((acc, request) => append({
    requests: [request],
    fetchable: pick(['url', 'method', 'body', 'query', 'headers'], request),
  }, acc), requests, std);

  map(async (block) => {
    map((request) => {
      // Update the status of requests in this block
      dispatch(setRequestStatus(request.hash, STATUS_INFLIGHT));
    }, prop('requests', block));

    try {
      let response, result;

      switch (method) {
        case METHOD_DOWNLOAD:
          ({ response, result } = await download(prop('fetchable', block)));
          break;
        default:
          ({
            response,
            result,
          } = await fetchWrapper(merge({
            auth: getCredential(getState()),
            headers: getHeaders(getState()),
          }, prop('fetchable', block))));
          break;
      }

      const pickHeaders = pick(['x-wp-totalpages', 'x-wp-total', 'link']);

      switch (true) {
        case endsWith('wp/v2/multi', block.fetchable.url):
          block.responses = map((Lresponse) => merge(pick(['status', 'body'], propOr({}, 'data', Lresponse)), {
            headers: pickHeaders(zipObj(
              map(toLower, keys(defaultTo({}, path(['data', 'headers'], Lresponse)))),
              values(path(['data', 'headers'], Lresponse)),
            )),
          }), result);
          break;

        case method === METHOD_DOWNLOAD:
          block.responses = [{
            headers: {},
            body: result,
            status: prop('status', response),
          }];
          break;

        default:
          const headers = {};
          if (response.headers.forEach) {
            // Use forEach if available
            response.headers.forEach((value, name) => {
              headers[ name ] = value;
            });
          } else {
            // fall back to more modern iterator-driven approach
            for (const pair of response.headers.entries()) {
              headers[ pair[ 0 ] ] = pair[ 1 ];
            }
          }

          block.responses = [{
            headers: pickHeaders(headers),
            body: result,
            status: prop('status', response),
          }];
          break;
      }
    } catch (err) {
      if (err instanceof Error) {
        switch (err.type) {
          case ERROR_INTERNAL:
            map((request) => {
              warn(FETCH, `Due to an internal error ${request.url} is being dequeued`);
              dispatch(dequeueRequest(request.hash));
              dispatch(reject(request.hash, propOr(err.msg, 'originalError', err)));
            }, block.requests);

            return block;
        }
      }

      if (err instanceof TypeError || err instanceof ReferenceError) {
        warn(FETCH, 'Fetch internal error', err);
      }

      map((request) => dispatch(setRequestStatus(request.hash, STATUS_RETRY)), block.requests);
      return block;
    }

    map(({ request, fetchable, response }) => {
      switch (request.method) {
        case METHOD_READ:
        case METHOD_UPDATE:
        case METHOD_CREATE:
        case METHOD_UPLOAD:
        case METHOD_MULTI:
          dispatch(addObject(prop('body', response), pick(['ttl', 'expires'], request)));
          break;

        case METHOD_DELETE:
          if (pathEq(['body', 'deleted'], true, response) && path(['body', 'previous'], response)) {
            dispatch(dropObject(prop('url', request)));
            break;
          }

          dispatch(addObject(prop('body', response), pick(['ttl', 'expires'], request)));
          break;

        case METHOD_DOWNLOAD:
          dispatch(addAsset(prop('url', request), merge(prop('body', response), pick(['ttl'], request))));
          break;
      }

      switch (drop(length(endpoint), request.url)) {
        case AUTH_ME_ENDPOINT:
          switch (response.status) {
            case HTTP_OK:
              dispatch(loggedIn(response.body));
              break;

            case HTTP_UNAUTHORIZED:
              dispatch(loggedOut());
              break;
          }
          break;
      }

      switch (response.status) {
        case HTTP_OK:
        case HTTP_CREATED:
          response.body = maybeMap((obj) => {
            const self = path(['_links', 'self', 0, 'href'], obj);

            if (!self) {
              return obj;
            }

            const cachedObject = getObjectBySelf(self, getState());

            if (!cachedObject) {
              return obj;
            }

            return {
              _self: self,
            };
          }, response.body);

          dispatch(updateRequest(request.hash, { response, status: STATUS_SUCCESS, now: Date.now() }));
          dispatch(resolve(prop('hash', request), response));
          break;

        default:
          dispatch(updateRequest(request.hash, { response, status: STATUS_FAILED, now: Date.now() }));
          dispatch(reject(prop('hash', request), response));
          break;
      }
    }, zipWith((request, response) => ({
      request,
      response,
    }), prop('requests', block), prop('responses', block)));

    return block;
  }, requests);

  return next(action);
};

export const reducer = (state = {}, action) => {
  switch (action.type) {
    case FETCH_SET:
      return merge(state, action.meta);

    case FETCH_SET_HEADER:
      return assocPath(['headers', action.name], action.value, state);
  }

  return state;
};
