import { compose } from 'ramda';
import { applyMiddleware } from 'redux';
import { createLogger } from 'redux-logger';

import { middlewares as analyticsMiddlewares } from './analytics';
import { middleware as assetMiddleware } from './asset';
import { middleware as authMiddleware } from './auth';
import { middleware as autoupdateMiddleware } from './autoupdate';
import { middleware as connectionMiddleware } from './connection';
import { middleware as eventMiddleware } from './event';
import { middleware as fetchMiddleware } from './fetch';
import { middleware as kvMiddleware } from './kv';
import { middleware as notificationMiddleware } from './notification';
import { middleware as objectMiddleware } from './object';
import { middleware as promiseMiddleware } from './promise';
import { middleware as queueMiddleware } from './queue';
import { middleware as rebootMiddleware } from './reboot';
import { middleware as schemaMiddleware } from './schema';
import { middleware as splashscreenMiddleware } from './splashscreen';
import { debounce, externalFlag } from './utils';
import { middleware as versionMiddleware } from './version';

/**
 * Get the required middleware for an external store
 */
const middlewares = [
  ...analyticsMiddlewares,
  autoupdateMiddleware,
  kvMiddleware,
  promiseMiddleware,
  queueMiddleware,
  fetchMiddleware,
  assetMiddleware,
  objectMiddleware,
  authMiddleware,
  notificationMiddleware,
  connectionMiddleware,
  rebootMiddleware,
  schemaMiddleware,
  splashscreenMiddleware,
  versionMiddleware,
  eventMiddleware,
];

if (externalFlag('redux')) {
  middlewares.push(createLogger({ collapsed: () => true }));
}

export const middleware = compose(
  applyMiddleware(...middlewares),
  batchedSubscribe(debounce((notify) => notify(), 0)),
);

// noinspection JSUnusedGlobalSymbols
export default middlewares;

// https://github.com/tappleby/redux-batched-subscribe
function batchedSubscribe(batch) {
  // noinspection UnnecessaryLocalVariableJS
  let currentListeners = [];
  let nextListeners = currentListeners;

  function ensureCanMutateNextListeners() {
    if (nextListeners === currentListeners) {
      nextListeners = currentListeners.slice();
    }
  }

  function subscribe(listener) {
    let isSubscribed = true;

    ensureCanMutateNextListeners();
    nextListeners.push(listener);

    return function unsubscribe() {
      if (!isSubscribed) {
        return;
      }

      isSubscribed = false;

      ensureCanMutateNextListeners();
      const index = nextListeners.indexOf(listener);
      nextListeners.splice(index, 1);
    };
  }

  function notifyListeners() {
    const listeners = currentListeners = nextListeners;
    for (let i = 0; i < listeners.length; i++) {
      listeners[ i ]();
    }
  }

  function notifyListenersBatched() {
    batch(notifyListeners);
  }

  return (next) => (...args) => {
    const store = next(...args);
    const subscribeImmediate = store.subscribe;

    function dispatch(...dispatchArgs) {
      const res = store.dispatch(...dispatchArgs);
      notifyListenersBatched();
      return res;
    }

    return {
      ...store,
      dispatch,
      subscribe,
      subscribeImmediate,
    };
  };
}
