import { contains, defaultTo, map, mapObjIndexed, max, merge, omit, path, prop, reject } from 'ramda';

import { EVENT_DISPATCH, OBJECT_ADD, OBJECT_DROP, QUEUE_PROCESS, STANDARD_TTL } from './const';
import { getExpiredObjects } from './selector';

export const add = (response, options) => ({
  type: OBJECT_ADD,
  response,
  options,
});

export const drop = (self) => ({
  type: OBJECT_DROP,
  self,
});

export const middleware = ({ dispatch, getState }) => (next) => (action) => {
  switch (action.type) {
    case QUEUE_PROCESS:
      mapObjIndexed((object, self) => dispatch(drop(self)), getExpiredObjects(getState()));
      break;
  }

  return next(action);
};

export const reducer = (state = {}, action) => {
  switch (action.type) {
    case OBJECT_ADD:
      let objects = defaultTo([], prop('response', action));
      const ttl = defaultTo(STANDARD_TTL, path(['options', 'ttl'], action));
      const expires = defaultTo([], path(['options', 'expires'], action));

      if (!Array.isArray(objects)) {
        objects = [objects];
      }

      map((object) => {
        const self = path(['_links', 'self', 0, 'href'], object);

        if (!self) {
          return;
        }

        const currentObject = defaultTo({}, prop(self, state));
        const currentTtl = defaultTo(0, prop('_ttl', currentObject));
        const newTtl = (ttl === false || currentTtl === false) ? false : max(ttl, currentTtl);

        if (JSON.stringify(omit(['_now', '_ttl', '_expires'], currentObject)) === JSON.stringify(object)) {
          state[ self ]._ttl = newTtl;
          state[ self ]._now = Date.now();
          state[ self ]._expires = expires;
          return;
        }

        // Update the current fetch timestamp of the object
        // Ensure options are correctly copied from the previous object
        state[ self ] = merge(object, {
          _now: Date.now(),
          _ttl: newTtl,
          _expires: expires,
        });
      }, objects);

      return state;

    case OBJECT_DROP:
      delete state[ action.self ];
      return state;

    case EVENT_DISPATCH:
      return reject((object) => contains(action.event, defaultTo([], object._expires)), state);
  }

  return state;
};
