import { compose, mapObjIndexed, mergeRight, path, replace, values } from 'ramda';
import { googlePlatform } from '../../analytics/google';
import { ANALYTICS_GOOGLE_PLATFORM as GOOGLE, ANALYTICS_TYPE } from '../../const';
import { stripUndefinedProps } from '../../utils';
import { getVisitId } from '../../analytics';

const insertVisitIdCustomDimension = (googleDimensions, visitId) =>
  mergeRight(googleDimensions || {}, { dimension1: visitId });

/**
 * Google Analytics platform-specific support
 * https://developers.google.com/analytics/devguides/collection/analyticsjs/field-reference
 */

/**
 * Google Analytics tracker initialization
 * @param {object} api - API object
 * @param {string} api.gaId - Google Analytics tracker ID
 * @param {object} api.appName - Application name
 * @param {object} api.appVersion - Application version
 * @returns {void}
 */
export const gaInit = ({ gaId, appName, appVersion }) => {
  // Validate a Google Analytics ID https://stackoverflow.com/questions/20411767#20412153
  if (!gaId || !/(UA|YT|MO)-\d+-\d+/i.test(gaId)) {
    return;
  }

  switch (googlePlatform()) {
    case GOOGLE.CORDOVA:
      window.ga.startTrackerWithId(gaId);
      window.ga.setAppVersion(appVersion);
      window.ga.setAnonymizeIp(true);
      window.ga.setVarP = (key, value) =>
        new Promise((resolve, reject) => window.ga.setVar(key, value, resolve, reject));
      break;

    case GOOGLE.TAG_MANAGER:
      const config = {
        app_name: appName,
        send_page_view: false,
        anonymize_ip: true,
        custom_map: {},
      };

      // Create simple mappings for custom dimensions and metrics
      for (let i = 0; i < 200; i++) {
        config.custom_map[ `dimension${i}` ] = `dimension${i}`;
        config.custom_map[ `metric${i}` ] = `metric${i}`;
      }

      window.gtag('config', gaId, config);
      break;
  }
};

/**
 * Helper function to set custom dimensions in the cordova plugin
 * @param dimensions
 * @returns {Promise<any[]>}
 */
export const cordovaSetCustomDimensions = (dimensions) =>
  Promise.all(values(mapObjIndexed(
    (val, key) => window.ga.setVarP(
      compose(replace('dimension', 'cd'), replace('contentGroup', 'cg'))(key),
      val,
    ),
    dimensions,
  )));

/**
 * Async function to call GA for page views, potentially waiting for data retrieval
 * @param {object} api - API object
 * @param {string} api.gaId - Google Analytics tracker ID
 * @param {object} api.site - Site object
 * @param {object} api.action - Request action
 * @returns {Promise<void>}
 */
export const gaPageView = async ({ gaId, site, action }) => {
  const { resource, objectRoute } = action;
  let { title, google } = action;

  let contentGroup = null;

  // If an object route was supplied, more information can be sent
  const object = site.getObjectByRoute(objectRoute);

  if (object) {
    if (!title) {
      title = path(['title', 'rendered'], object);
    }

    // Group content on GA on our object type
    contentGroup = `/${object.type}`;
  }

  google = insertVisitIdCustomDimension(google, getVisitId(site.store.getState()));

  switch (googlePlatform()) {
    case GOOGLE.CORDOVA:
      if (contentGroup) {
        await window.ga.setVarP('cg1', contentGroup);
      }

      await cordovaSetCustomDimensions(google);

      window.ga.trackView(title, `storyshare://${resource}`);
      window.ga.dispatch();
      break;

    case GOOGLE.TAG_MANAGER:
      const event = {
        page_title: title,
        page_path: resource,
        ...google,
      };

      window.gtag('config', gaId, stripUndefinedProps(event));
      break;
  }
};

/**
 * Async function to call GA for events, potentially waiting for data retrieval
 * @param {object} api - API object
 * @param {object} api.site - Site object
 * @param {object} api.action - Request action
 * @returns {Promise<void>}
 */
export const gaEvent = async ({ site, action }) => {
  const {
    interactive = true,
    eventType,
    action: eventAction,
    value: eventValue,
    objectRoute,
  } = action;

  let { label: eventLabel, google } = action;

  if (!eventLabel) {
    const object = site.getObjectByRoute(objectRoute);

    if (object) {
      eventLabel = path(['title', 'rendered'], object);
    }
  }

  const typeMapping = {
    [ ANALYTICS_TYPE.GENERIC ]: 'none',
    [ ANALYTICS_TYPE.APP ]: 'app',
    [ ANALYTICS_TYPE.AUTH ]: 'auth',
    [ ANALYTICS_TYPE.NOTIFICATION ]: 'notification',
    [ ANALYTICS_TYPE.MEDIA ]: 'media',
    [ ANALYTICS_TYPE.SOCIAL ]: 'social',
    [ ANALYTICS_TYPE.SEARCH ]: 'search',
    [ ANALYTICS_TYPE.NAVIGATION ]: 'navigation',
  };

  const eventCategory = typeMapping[ eventType ] || 'none';

  google = insertVisitIdCustomDimension(google, getVisitId(site.store.getState()));

  switch (googlePlatform()) {
    case GOOGLE.CORDOVA:
      await cordovaSetCustomDimensions(google);

      window.ga.trackEvent(eventCategory, eventAction, eventLabel, eventValue);
      window.ga.dispatch();
      break;

    case GOOGLE.TAG_MANAGER:
      const event = {
        event_category: eventCategory,
        event_label: eventLabel,
        value: eventValue,
        non_interaction: interactive === false,
        ...google,
      };

      window.gtag('event', eventAction, stripUndefinedProps(event));
      break;
  }
};
