import 'cross-fetch/polyfill';
import URI from 'urijs';

import { authenticate, setStatus } from '../auth';
import {
  AUTH,
  AUTH_STATUS_LOGGED_IN,
  AUTH_STATUS_PENDING,
  AUTH_STATUS_RETRY,
  ERROR_INTERNAL,
  EVENT_AUTH_LOGGED_IN,
  PLATFORM_CORDOVA,
  SECOND,
} from '../const';
import { debug } from '../debug';
import Error from '../error';
import { generateEvent } from '../event';
import { get as getMeta } from '../meta';
import { Promise } from '../promise';
import { platform } from '../utils';

/**
 * Attempts to log in automatically
 * @param {Store} store - Redux store
 * @returns {Bluebird<any>}
 */
export const loginAutomatic = ({ dispatch, getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      dispatch(setStatus(AUTH_STATUS_PENDING));

      // Set up a function to handle the Cordova callback
      const endpoint = getMeta(getState(), 'endpoint');
      const authURI = URI(endpoint).path('/storyshare/appauthenticate/');

      return new Promise((resolve, reject) => {
        // Open the authentication processor in the external browser

        debug(AUTH, 'Opening inappbrowser in hidden mode');

        // noinspection JSUnresolvedVariable
        let ref = cordova.InAppBrowser.open(authURI.toString(), '_blank', 'location=no,hidden=yes,toolbar=no');

        ref.addEventListener('loadstop', () => {
          let injection = false;

          // noinspection JSUnresolvedFunction
          ref.executeScript({ code: 'window.__storyshare_authentication_key__' }, (values) => {
            injection = true;

            const key = values[ 0 ];

            if (!key) {
              ref.show();
              return;
            }

            authURI.query({ key });

            resolve(fetch(authURI.toString(), { credentials: 'include' }));

            ref.close();
            ref = undefined;
          });

          // If injection fails (due to incompatible CSP), show the IAB
          setTimeout(() => {
            if (injection) {
              return;
            }

            ref.show();
          }, SECOND / 2);
        });

        ref.addEventListener('loaderror', reject);
      })
        .then((response) => {
          if (response.status !== 200) {
            throw new Error(ERROR_INTERNAL, 'Did not get a 200 response from the server post-auth', response);
          }

          dispatch(setStatus(AUTH_STATUS_LOGGED_IN));
          dispatch(generateEvent(EVENT_AUTH_LOGGED_IN));

          return null;
        })
        .catch(() => {
          dispatch(setStatus(AUTH_STATUS_RETRY));

          return Promise.delay(2 * SECOND).then(() => dispatch(authenticate()));
        });

    default:
      // Cannot authenticate on this platform
      return Promise.reject(
        new Error(ERROR_INTERNAL, `Cannot automatically authenticate on the ${platform()} platform`),
      );
  }
};
