import 'cross-fetch/polyfill';
import { defaultTo, forEach, isEmpty, prop, propEq, values } from 'ramda';
import URI from 'urijs';

import {
  getActive,
  getHashes,
  getLoad,
  getRepository,
  setActive,
  setLastChecked,
  setLoad,
  setProgress,
  setStatus,
  storeHashes,
} from '../autoupdate';
import {
  AUTOUPDATE,
  AUTOUPDATE_STATUS_ERROR,
  AUTOUPDATE_STATUS_IDLE,
  AUTOUPDATE_STATUS_PENDING_RESTART,
  AUTOUPDATE_STATUS_UPDATING,
  ERROR_INTERNAL,
  HTTP_OK,
  PLATFORM_CORDOVA,
  VERSION_AUTOUPDATE,
} from '../const';
import { debug, error, warn } from '../debug';
import Error from '../error';
import { get as getMeta } from '../meta';
import { Promise } from '../promise';
import { reboot } from '../reboot';
import { deleteDirectory, fileExists, readFile, recursiveCopy, writeFile } from '../storage';
import { platform } from '../utils';
import { set as setVersion } from '../version';

/**
 * Checks whether an update is available and installs it if possible
 * @param {MiddlewareAPI} api - Redux store methods
 * @returns {Promise}
 */
export const check = async ({ dispatch, getState }) => {
  const state = getState();

  switch (platform()) {
    case PLATFORM_CORDOVA:
      const timer = Date.now();
      let currentDirectory = getActive(state);
      const nextDirectory = Math.random().toString();

      try {
        dispatch(setProgress(0));
        dispatch(setStatus(AUTOUPDATE_STATUS_UPDATING));

        debug(AUTOUPDATE, 'Requesting manifest');

        const site = getMeta(state, 'site');
        const manifest = await site.req(`storyshare/v1/repository/${getRepository(state)}/manifest`).promise().read();
        dispatch(setLastChecked());

        // Delete the next directory if it exists
        await deleteDirectory(nextDirectory);

        // Get the list of hashes from the store
        let hashes = defaultTo({}, getHashes(state));

        // If the list of hashes is empty then unset the current directory
        // and get the hashes from the compiled version
        if (isEmpty(hashes)) {
          debug(AUTOUPDATE, 'No hashes exist in the store, getting hashes from builtin');
          currentDirectory = null;

          try {
            hashes = JSON.parse(
              String.fromCharCode.apply(
                null,
                new Uint8Array(
                  await readFile('www/manifest.json', cordova.file.applicationDirectory),
                ),
              ),
            );

            dispatch(storeHashes(hashes));
          } catch (err) {
            warn(AUTOUPDATE, 'No built-in manifest was found');
          }
        }

        debug(AUTOUPDATE, 'Determining items to download');

        const downloadList = await Promise.filter(values(manifest.files), async (manifestEntry, index, length) => {
          dispatch(setProgress(((length - index - 1) / length) * 50));

          const { filename, version } = manifestEntry;

          // Skip the index.html (it should not exist in the manifest)
          if (filename === 'index.html') {
            return false;
          }

          return !propEq(filename, version, hashes);
        }, { concurrency: 1 });

        if (!isEmpty(downloadList)) {
          // Copy either the compiled in version or the current version
          if (currentDirectory) {
            debug(AUTOUPDATE, `Copying ${currentDirectory} to ${nextDirectory}`);
            await recursiveCopy(currentDirectory, nextDirectory);
          } else {
            debug(AUTOUPDATE, `Copying builtin directory to ${nextDirectory}`);
            await recursiveCopy('www', nextDirectory, cordova.file.applicationDirectory);
          }

          debug(AUTOUPDATE, `Downloading ${downloadList.length} items`);

          await Promise.mapSeries(downloadList, async (file, index, length) => {
            const { url, filename, version } = file;

            dispatch(setProgress(((index / length) * 50) + 50));

            const targetFile = `${nextDirectory}/${filename}`;
            const response = await fetch(url);

            // noinspection JSUnresolvedVariable
            if (response.status !== HTTP_OK) {
              throw new Error(ERROR_INTERNAL, 'Unable to complete download');
            }

            // noinspection JSUnresolvedFunction
            const buffer = await response.arrayBuffer();

            // noinspection JSUnresolvedFunction
            await writeFile(targetFile, buffer);
            hashes[ filename ] = version;
          });

          debug(AUTOUPDATE, 'Completed download');

          dispatch(storeHashes(hashes));
          dispatch(setStatus(AUTOUPDATE_STATUS_PENDING_RESTART));
          dispatch(setVersion(VERSION_AUTOUPDATE, prop('lastUpdated', manifest)));
          dispatch(setLoad(prop('load', manifest)));
          dispatch(setActive(nextDirectory));

          if (currentDirectory) {
            await deleteDirectory(currentDirectory);
          }
        } else {
          dispatch(setStatus(AUTOUPDATE_STATUS_IDLE));
        }

        debug(AUTOUPDATE, `Completed autoupdate in ${Date.now() - timer} ms`);
      } catch (err) {
        error(AUTOUPDATE, 'Autoupdate system encountered an error', err);
        dispatch(setStatus(AUTOUPDATE_STATUS_ERROR));
        await deleteDirectory(nextDirectory);
      }

      dispatch(setProgress(100));
  }
};

/**
 * Initialise this platform for autoupdate
 * @param api
 * @returns {Promise<void>}
 */
export const init = async ({ dispatch, getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      const activeDirectory = getActive(getState());

      const target = `${activeDirectory}/index.html`;

      // If the active version exists but our current path does not match the active version then
      // set pending restart status
      if (await fileExists(target) && !URI().path().endsWith(target)) {
        dispatch(setStatus(AUTOUPDATE_STATUS_PENDING_RESTART));
      }
  }
};

/**
 * Computes the location of autoupdated content and reboots to load it
 * @param {MiddlewareAPI} api - Redux store methods
 * @returns {Promise<boolean>} whether an update took place
 */
export const redirect = async ({ dispatch, getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      const activeDirectory = getActive(getState());

      const target = `${activeDirectory}/index.html`;

      // Last minute check that the target exists
      if (!await fileExists(target)) {
        return false;
      }

      return dispatch(reboot(`${cordova.file.dataDirectory}/${target}`));

    default:
      return false;
  }
};

/**
 * Loads any over-the-top assets provided by the repository such as extra javascript or css files
 * @param {array} assets - List of asset filenames
 */
export const loadAssets = ({ getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      const activeDirectory = getActive(getState());

      forEach((asset) => {
        switch (URI(asset).suffix()) {
          case 'js':
            const script = document.createElement('script');
            script.type = 'text/javascript';
            script.src = `${cordova.file.dataDirectory}/${activeDirectory}/${asset}`;
            document.head.appendChild(script);
            break;

          case 'css':
            const link = document.createElement('link');
            link.type = 'text/css';
            link.rel = 'stylesheet';
            link.href = `${cordova.file.dataDirectory}/${activeDirectory}/${asset}`;
            document.head.appendChild(link);
            break;
        }
      }, getLoad(getState()));
  }
};
