import fs from 'fs';

import mzfs from 'mz/fs';
import path from 'path';
import { merge, prop, startsWith } from 'ramda';
import URI from 'urijs';

import { ERROR_INTERNAL, ERROR_USAGE, PLATFORM_BROWSER, PLATFORM_CORDOVA, PLATFORM_NODEJS } from '../const';
import Error from '../error';
import { readFile } from '../storage';
import { isFile, platform } from '../utils';

/**
 * Modifies fetch options for upload, loading and attaching a file
 * @param {any} body - option passed to `fetchWrapper()`
 * @param {RequestOptions } options - request options
 * @returns {void}
 */
export const prepareForUpload = async (body, options) => {
  let filename = null;

  switch (platform()) {
    case PLATFORM_NODEJS:
      filename = prop('file', body);

      let stats = null;
      try {
        stats = await mzfs.stat(filename);
      } catch (err) {
        throw new Error(ERROR_INTERNAL, `Could not read file at ${filename}`, err);
      }

      options.body = fs.createReadStream(filename);
      options.headers = merge(options.headers, {
        'Content-Length': stats.size,
        'Content-Type': 'application/octet-stream',
        'Content-Disposition': `attachment; filename="${path.basename(filename)}"`,
      });
      break;

    case PLATFORM_BROWSER:
    case PLATFORM_CORDOVA:
      const file = prop('file', body);

      switch (true) {
        case isFile(file):
          filename = file.name;
          options.body = file;
          break;

        case platform() === PLATFORM_CORDOVA && file && startsWith('file://', file):
          filename = URI(file).filename();

          try {
            options.body = await readFile(file);
          } catch (err) {
            throw new Error(ERROR_INTERNAL, 'Failed to read file');
          }
          break;

        default:
          throw new Error(ERROR_USAGE, 'Could not understand supplied file');
      }

      options.headers[ 'Content-Disposition' ] = `attachment; filename="${filename}"`;
      break;
  }
};
