import { propOr } from 'ramda';

import { event as trackEvent } from '../analytics';
import {
  ANALYTICS_TYPE,
  DAY,
  ERROR_INTERNAL,
  EVENT_NOTIFICATION_RECEIVED,
  NOTIFICATION,
  PLATFORM_CORDOVA,
  PLATFORM_CORDOVA_ANDROID,
  PLATFORM_CORDOVA_IOS,
  PLATFORM_CORDOVA_UNKNOWN,
} from '../const';
import { debug, error, warn } from '../debug';
import Error from '../error';
import { generateEvent } from '../event';
import { get as getMeta } from '../meta';
import { register as registerAction, setEnabled } from '../notification';
import { externalFlag, platform } from '../utils';

/**
 * Global push notification instance on the Cordova platform
 * @type {PhonegapPluginPush.PushNotification | null}
 */
let cordovaPushInstance = null;

/**
 * Helper to determine the device in use on the Cordova platform
 * @returns {PLATFORM_CORDOVA_ANDROID | PLATFORM_CORDOVA_IOS | PLATFORM_CORDOVA_UNKNOWN}
 */
export const cordovaPlatform = () =>
  platform() === PLATFORM_CORDOVA
    ? window.cordova.platformId
    : PLATFORM_CORDOVA_UNKNOWN;

/**
 * Refreshes notification status
 * @param {MiddlewareAPI} api - Redux store methods
 * @returns {void}
 */
export const refresh = ({ dispatch }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      if (!window.PushNotification) {
        warn(NOTIFICATION, 'Push notification plugin is not installed');
        break;
      }

      window.PushNotification.hasPermission((data) => {
        dispatch(setEnabled(propOr(false, 'isEnabled', data)));

        if (data.isEnabled) {
          dispatch(registerAction());
        }
      });
      break;
  }
};

/**
 * Registers for notifications
 * @param {MiddlewareAPI} api - Redux store methods
 * @returns {void}
 */
export const register = ({ dispatch, getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      cordovaPushInstance = window.PushNotification.init({
        android: {},
        ios: {
          'sound': true,
          'vibration': true,
          'badge': true,
        },
        windows: {},
      });

      cordovaPushInstance.on('error', (e) => {
        error(NOTIFICATION, 'Received error from the push notification service', e);
      });

      let type = externalFlag('notificationType');

      if (!type) {
        switch (cordovaPlatform()) {
          case PLATFORM_CORDOVA_IOS:
            type = 'apn';
            break;

          case PLATFORM_CORDOVA_ANDROID:
            type = 'gcm';
            break;
        }
      }

      if (!type) {
        throw new Error(ERROR_INTERNAL, 'Could not determine Cordova platform type');
      }

      cordovaPushInstance.on('registration', (registration) => {
        debug(NOTIFICATION, 'Device registered for push notifications');

        const site = getMeta(getState(), 'site');

        site.req('storyshare/v1/notification/register')
          .ttl(DAY)
          .retry(DAY)
          .update({
            type,
            key: registration.registrationId,
          });
      });

      cordovaPushInstance.on('notification', (notification) => {
        debug(NOTIFICATION, 'Received notification', notification);
        dispatch(generateEvent(EVENT_NOTIFICATION_RECEIVED, notification));
        dispatch(trackEvent(ANALYTICS_TYPE.NOTIFICATION, { action: 'received', interactive: false }));
      });
  }
};

/**
 * Cancels registration for notifications
 * @param {MiddlewareAPI} api - Redux store methods
 */
export const unregister = ({ dispatch }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      if (typeof cordovaPushInstance === 'object') {
        cordovaPushInstance.unregister();
      }

      dispatch(setEnabled(false));
      break;
  }
};
