import hash from 'object-hash';
import { startsWith } from 'ramda';
import URI from 'urijs';

import { Promise } from '../../promise';
import { stripQuery } from '../../utils';

// Promisify a single Cordova function
const cordovaPromisify = (fn) => (...args) => new Promise((resolve, reject) => fn(...args, resolve, reject));

// Promisify a Cordova object
const cordovaPromisifyObject = (obj) => Promise.promisifyAll(obj, {
  suffix: 'P',
  promisifier: (m) => function (...args) {
    return new Promise((resolve, reject) => m.call(this, ...args, resolve, reject));
  },
});

const resolveLocalFileSystemURL = (root) => cordovaPromisify(window.resolveLocalFileSystemURL)(root);

const createNestedDirectory = async (path, root = cordova.file.dataDirectory) =>
  Promise.reduce(path.split('/'), (de, p) =>
    cordovaPromisifyObject(de).getDirectoryP(p, { create: true }), await resolveLocalFileSystemURL(root));

const resolveFileEntry = async (path, root) => {
  if (startsWith('file://', path)) {
    return resolveLocalFileSystemURL(path);
  }

  const fs = await resolveLocalFileSystemURL(root);
  return cordovaPromisifyObject(fs).getFileP(path, {});
};

export const readFile = async (path, root = cordova.file.dataDirectory) => {
  const fileEntry = await resolveFileEntry(path, root);
  const file = await cordovaPromisifyObject(fileEntry).fileP();

  return new Promise((resolve, reject) => {
    const reader = new FileReader();
    reader.onloadend = (e) => resolve(e.target.result);
    reader.onerror = (e) => reject(e.target);
    reader.readAsArrayBuffer(file);
  });
};

export const rootDir = () => cordova.file.dataDirectory;

export const writeFile = async (path, data, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  let directory = path.split('/');
  directory.pop();
  directory = directory.join('/');

  await createNestedDirectory(directory, root);

  const fe = await cordovaPromisifyObject(fs).getFileP(path, { create: true });
  const fw = await cordovaPromisifyObject(fe).createWriterP();

  return new Promise((resolve, reject) => {
    fw.onwriteend = (e) => resolve(e.target);
    fw.onerror = (e) => reject(e.target);
    fw.write(data);
  });
};

export const fileExists = async (path, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  try {
    await cordovaPromisifyObject(fs).getFileP(path, {});
    return true;
  } catch (err) {
    return false;
  }
};

export const directoryExists = async (path, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  try {
    await cordovaPromisifyObject(fs).getDirectoryP(path, {});
    return true;
  } catch (err) {
    return false;
  }
};

export const recursiveCopy = async (
  from,
  to,
  fromRoot = cordova.file.dataDirectory,
  toRoot = cordova.file.dataDirectory,
) => {
  const fromFs = cordovaPromisifyObject(await resolveLocalFileSystemURL(fromRoot));
  const toFs = cordovaPromisifyObject(await resolveLocalFileSystemURL(toRoot));

  try {
    const d = await toFs.getDirectoryP(to, {});
    await cordovaPromisifyObject(d).removeRecursivelyP();
    // eslint-disable-next-line no-empty
  } catch (err) {
  }

  for (const f of await cordovaPromisifyObject((await fromFs.getDirectoryP(from, {})).createReader()).readEntriesP()) {
    await cordovaPromisifyObject(f).copyToP(await createNestedDirectory(to, toRoot), undefined);
  }
};

export const moveDirectory = async (from, to, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  const fromDe = await cordovaPromisifyObject(fs).getDirectoryP(from, {});

  let p = to.split('/');
  const t = p.pop();
  p = p.join('/');
  const toDe = await cordovaPromisifyObject(fs).getDirectoryP(p, { create: true });

  try {
    await cordovaPromisifyObject(fromDe).moveToP(toDe, t);
    // eslint-disable-next-line no-empty
  } catch (err) {
  }
};

export const deleteDirectory = async (path, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  try {
    const de = await cordovaPromisifyObject(fs).getDirectoryP(path, {});
    await cordovaPromisifyObject(de).removeRecursivelyP();
    // eslint-disable-next-line no-empty
  } catch (err) {
  }
};

export const deleteFile = async (path, root = cordova.file.dataDirectory) => {
  const fs = await resolveLocalFileSystemURL(root);

  try {
    const fe = await cordovaPromisifyObject(fs).getFileP(path, {});
    await cordovaPromisifyObject(fe).removeP();
    // eslint-disable-next-line no-empty
  } catch (err) {
  }
};

export const downloadFile = async (url, root = cordova.file.dataDirectory) => {
  const filename = `${hash(stripQuery(url))}.${URI(url).suffix()}`;

  try {
    const fileTransfer = new window.FileTransfer();
    await createNestedDirectory('assets', root);

    // attempt to download file and place it in assets folder
    await new Promise((resolve, reject) =>
      fileTransfer.download(url, `${root}assets/${filename}`, resolve, reject)
    );

    return filename;
    // eslint-disable-next-line no-empty
  } catch (err) {
  }
};
