import localforage from 'localforage';
import fs from 'mz/fs';
import os from 'os';
import fsPath from 'path';

import { ERROR_INTERNAL, PLATFORM_CORDOVA, PLATFORM_NODEJS } from '../../const';
import Error from '../../error';
import { platform } from '../../utils';
import {
  deleteDirectory as cordovaDeleteDirectory,
  deleteFile as cordovaDeleteFile,
  directoryExists as cordovaDirectoryExists,
  fileExists as cordovaFileExists,
  moveDirectory as cordovaMoveDirectory,
  readFile as cordovaReadFile,
  recursiveCopy as cordovaRecursiveCopy,
  rootDir as cordovaRootDir,
  writeFile as cordovaWriteFile,
} from './cordova';
import {
  deleteFile as nodeDeleteFile,
  readFile as nodeReadFile,
  rootDir as nodeRootDir,
  writeFile as nodeWriteFile,
} from './nodejs';

const FILE = 'storyshare.json';

/**
 * Whether a save is currently in progress
 * @type {boolean}
 */
let writeLock = false;

/**
 * Loads store data from persistent storage
 * @returns {Promise<string>} store data
 */
export const load = async () => {
  switch (platform()) {
    case PLATFORM_NODEJS:
      const file = fsPath.join(os.tmpdir(), FILE);

      if (!await fs.exists(file)) {
        return null;
      }

      return fs.readFile(file, 'utf8');

    default:
      return localforage.getItem('storyshare');
  }
};

/**
 * Saves store data to persistent storage
 * @param {string} store - store data
 */
export const save = (store) => {
  switch (platform()) {
    case PLATFORM_NODEJS:
      // Check if we can get the write lock
      if (writeLock) {
        return;
      }

      const file = fsPath.join(os.tmpdir(), FILE);
      // Obtain the write lock
      writeLock = true;
      fs.writeFile(file, store, 'utf8', () => {
        // release write lock
        writeLock = false;
      });
      break;

    default:
      localforage.setItem('storyshare', store);
  }
};

export const rootDir = () => {
  switch (platform()) {
    case PLATFORM_NODEJS:
      return nodeRootDir();

    case PLATFORM_CORDOVA:
      return cordovaRootDir();

    default:
      return '';
  }
};

export const readFile = async (path, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaReadFile(path, root);

    case PLATFORM_NODEJS:
      return nodeReadFile(path);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const writeFile = async (path, data, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaWriteFile(path, data, root);

    case PLATFORM_NODEJS:
      return nodeWriteFile(path, data);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const fileExists = async (path, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaFileExists(path, root);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const directoryExists = async (path, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaDirectoryExists(path, root);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const moveDirectory = async (from, to, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaMoveDirectory(from, to, root);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const deleteDirectory = async (path, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaDeleteDirectory(path, root);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const recursiveCopy = async (from, to, fromRoot, toRoot) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaRecursiveCopy(from, to, fromRoot, toRoot);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};

export const deleteFile = async (path, root) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      return cordovaDeleteFile(path, root);

    case PLATFORM_NODEJS:
      return nodeDeleteFile(path);

    default:
      throw new Error(ERROR_INTERNAL, 'Not implemented');
  }
};
