import { contains } from 'ramda';

import { OFFLINE, ONLINE, PAUSE, PLATFORM_BROWSER, PLATFORM_CORDOVA, RESUME, UNLOAD } from '../const';
import { Promise } from '../promise';
import { platform } from '../utils';

/**
 * Retrieves extra platform-specific store enhancers
 * @returns {StoreEnhancer[]}
 */
export const getEnhancers = () => {
  if (platform() === PLATFORM_BROWSER && window.__REDUX_DEVTOOLS_EXTENSION__) {
    return [window.__REDUX_DEVTOOLS_EXTENSION__()];
  } else {
    return [];
  }
};

/**
 * Waits until the platform has started up and is ready for use
 * @returns {Promise<void>}
 */
export const ready = () => {
  if (platform() === PLATFORM_CORDOVA) {
    return new Promise(
      (resolve) => document.addEventListener('deviceready', resolve, false),
    );
  } else {
    return Promise.resolve();
  }
};

/**
 * Adds platform-specific event listeners
 * @param {Dispatch<AnyAction>} dispatch - Redux store dispatch method
 * @returns {void}
 */
export const addEventListeners = (dispatch) => {
  const addListener = (event, type, object) =>
    object.addEventListener(event, () => dispatch({ type }), false);

  if (contains(platform(), [PLATFORM_BROWSER, PLATFORM_CORDOVA])) {
    addListener('resume', RESUME, document);
    addListener('pause', PAUSE, document);
    addListener('online', ONLINE, document);
    addListener('offline', OFFLINE, document);
    addListener('beforeunload', UNLOAD, window);
  }
};

/**
 * Checks whether the platform permits scheduled tasks
 * @returns {boolean}
 */
export const supportsScheduledTasks = () =>
  contains(platform(), [PLATFORM_BROWSER, PLATFORM_CORDOVA]);
