import { event as trackEvent } from '../analytics';
import {
  ANALYTICS_TYPE,
  CORDOVA,
  PLATFORM_CORDOVA,
  VERSION_APPCONTAINER,
  VERSION_AUTOUPDATE,
  VERSION_PACKAGENAME,
} from '../const';
import { warn } from '../debug';
import { platform } from '../utils';
import { getVersion, set } from '../version';

/**
 * Assembles an app version string for the current platform
 * @param {Store} state - store state
 * @returns {string|undefined}
 */
export const formatVersion = (state) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      let version = state.version[ VERSION_APPCONTAINER ];
      const autoupdateVersion = state.version[ VERSION_AUTOUPDATE ];

      if (autoupdateVersion) {
        version = `${version}.${autoupdateVersion}`;
      }

      return version;
  }
};

/**
 * Refreshes the app version, saving it in the store
 * @param {MiddlewareAPI} api - Redux store methods
 * @returns {Promise} that resolves when the version has been refreshed
 */
export const refresh = ({ dispatch, getState }) => {
  switch (platform()) {
    case PLATFORM_CORDOVA:
      if (!cordova.getAppVersion) {
        warn(CORDOVA, 'App version plugin is not installed');
        break;
      }

      return Promise.all([
        cordova.getAppVersion.getVersionNumber(),
        cordova.getAppVersion.getPackageName(),
      ])
        .then(([version, packageName]) => {
          dispatch(set(VERSION_APPCONTAINER, version));
          dispatch(set(VERSION_PACKAGENAME, packageName));

          // If the state did not have a previous version number, track this as an install
          if (!getVersion(VERSION_APPCONTAINER, getState())) {
            dispatch(trackEvent(ANALYTICS_TYPE.APP, { action: 'install', interactive: false }));
          }
        });
  }
};
