import {
  __,
  adjust,
  any,
  append,
  contains,
  defaultTo,
  equals,
  filter,
  findIndex,
  isEmpty,
  lt,
  map,
  merge,
  not,
  path,
  prop,
  propEq,
  reject,
  subtract,
} from 'ramda';

import {
  EVENT_DISPATCH,
  HTTP_OK,
  METHOD_READ,
  QUEUE_ADD,
  QUEUE_DEQUEUE,
  QUEUE_DROP,
  QUEUE_ENQUEUE,
  QUEUE_PROCESS,
  QUEUE_CLEAR_BACKOFF,
  QUEUE_UPDATE,
  RESUME,
  STANDARD_RETRY,
  STANDARD_TTL,
  STATUS_BACKOFF,
  STATUS_READY,
  STATUS_SUCCESS,
} from './const';
import { get as getMeta } from './meta';
import { add as addPromise, Promise } from './promise';
import Response from './response';
import {
  getExpiredRequests,
  getRefreshableRequests,
  getRequestByHash,
  getRequestsByTrackingKey,
  getRequestsByStatus,
} from './selector';
import { debounce, maybePromise } from './utils';

export const enqueueRequest = (request, options = {}) => ({
  type: QUEUE_ENQUEUE,
  request,
  options,
});

export const dequeueRequest = (hash) => ({
  type: QUEUE_DEQUEUE,
  hash,
});

export const setRequestStatus = (hash, status) => ({
  type: QUEUE_UPDATE,
  hash,
  delta: { status },
});

export const updateRequest = (hash, delta) => ({
  type: QUEUE_UPDATE,
  hash,
  delta,
});

const addRequest = (request) => ({
  type: QUEUE_ADD,
  request,
});

const dropRequest = (hash) => ({
  type: QUEUE_DROP,
  hash,
});

export const queueProcess = () => ({
  type: QUEUE_PROCESS,
});

export const runPipeline = debounce((dispatch) =>
  setTimeout(() => dispatch(queueProcess()), 0),
);

export const middleware = ({ getState, dispatch }) => (next) => (action) => {
  const site = getMeta(getState(), 'site');

  switch (action.type) {
    case QUEUE_ENQUEUE:
      const { request, options } = action;
      const { hash } = options;

      // Try and find an existing request that matches the hash of this one
      const existingRequest = getRequestByHash(hash, getState());

      // Generate request options from the existing request (if it exists), overridden with passed options from this
      // request, together with defaults if nothing was specified
      const {
        track,
        retry = site._options.defaultRetry || STANDARD_RETRY,
        soft = false,
        force = false,
        promise = false,
        ttl = defaultTo(STANDARD_TTL, site._options.defaultTtl),
        refresh = false,
        expires = [],
      } = merge(defaultTo({}, existingRequest), options);

      // If a tracking code is provided, remove it from any request that was using it (except this request)
      if (track) {
        map(
          (r) => dispatch(updateRequest(r.hash, { track: null })),
          filter(
            (r) => not(propEq('hash', hash, r)),
            getRequestsByTrackingKey(track, getState()),
          ),
        );

        // If an existing request existed, reattach the tracking key to it
        if (existingRequest) {
          dispatch(updateRequest(existingRequest.hash, { track }));
        }
      }

      // Check whether a soft response can be returned
      let softResponse = false;

      if ((typeof soft === 'boolean' && soft) || typeof soft === 'number') {
        // Try a soft response from a request
        if (existingRequest && !isEmpty(defaultTo({}, prop('response', existingRequest)))) {
          if (soft === true || lt(subtract(Date.now(), prop('now', existingRequest)), soft)) {
            softResponse = new Response(prop('response', existingRequest), site);
          }
        }

        // Try a soft response from a cached object
        let existingObject;
        if (
          !softResponse
          && request.method === METHOD_READ
          && isEmpty(request.query)
          && (existingObject = path(['objects', request.url], getState()))
        ) {
          if (soft === true || lt(subtract(Date.now(), prop('_now', existingObject)), soft)) {
            const response = {
              headers: {},
              status: HTTP_OK,
              body: {
                _self: request.url,
              },
            };

            // Add a backdated request so the soft response from an object can be tracked
            dispatch(addRequest(merge(request, {
              now: existingObject._now,
              hash,
              retry,
              track,
              ttl: existingObject._ttl,
              refresh,
              expires,
              response,
              status: STATUS_SUCCESS,
            })));

            softResponse = new Response(response, site);
          }
        }
      }

      if (softResponse && !force) {
        return maybePromise(softResponse, promise);
      }

      // If an existing request exists for this content, just set its status back to ready
      // Otherwise, create a new request
      if (existingRequest) {
        dispatch(updateRequest(existingRequest.hash, {
          query: request.query, // Query parameters related to fetching signed assets may have changed
          now: Date.now(),
          retry,
          track,
          ttl,
          refresh,
          expires,
          attempt: 1,
          status: STATUS_READY,
        }));
      } else {
        dispatch(addRequest(merge(request, {
          now: Date.now(),
          hash,
          track,
          retry,
          ttl,
          refresh,
          expires,
          status: STATUS_READY,
        })));
      }

      runPipeline(dispatch);

      if (softResponse) {
        return maybePromise(softResponse, promise);
      }

      // eslint-disable-next-line no-shadow
      return promise ? new Promise((resolve, reject) => dispatch(addPromise(hash, resolve, reject))) : null;

    case QUEUE_DEQUEUE:
      dispatch(dropRequest(prop('hash', action)));
      runPipeline(dispatch);
      break;

    case RESUME:
    case QUEUE_PROCESS:
      map((Lrequest) => dispatch(dropRequest(prop('hash', Lrequest))), getExpiredRequests(getState()));
      map((Lrequest) => dispatch(updateRequest(Lrequest.hash, {
        now: Date.now(),
        status: STATUS_READY,
      })), getRefreshableRequests(getState()));
      break;

    case QUEUE_CLEAR_BACKOFF:
      map((Lrequest) => dispatch(updateRequest(Lrequest.hash, {
        now: Date.now(),
        status: STATUS_READY,
      })), getRequestsByStatus(STATUS_BACKOFF, getState()));
      break;

    case QUEUE_UPDATE:
      runPipeline(dispatch);
      break;
  }

  return next(action);
};

export const reducer = (state = [], action) => {
  switch (action.type) {
    case QUEUE_ADD:
      if (any(propEq('hash', path(['request', 'hash'], action)), state)) {
        return state;
      }
      return append(prop('request', action), state);

    case QUEUE_DROP:
      return reject((request) => equals(prop('hash', request), prop('hash', action)), state);

    case QUEUE_UPDATE:
      return adjust(
        findIndex(propEq('hash', prop('hash', action)), state),
        merge(
          __,
          prop('delta', action),
        ),
        state,
      );

    case EVENT_DISPATCH:
      return reject((request) => contains(action.event, defaultTo([], request.expires)), state);
  }

  return state;
};
