import objectHash from 'object-hash';
import { append, assocPath, contains, defaultTo, has, merge, omit, path, uniq, without } from 'ramda';

import {
  ERROR_USAGE,
  METHOD_CREATE,
  METHOD_DELETE,
  METHOD_DOWNLOAD,
  METHOD_READ,
  METHOD_UPDATE,
  METHOD_UPLOAD,
  REQUEST,
} from './const';
import { warn } from './debug';
import Error from './error';
import * as platform from './platforms/request';
import { cancel as cancelPromise } from './promise';
import { enqueueRequest } from './queue';
import Response from './response';
import { httpMethod, isFile, maybePromise } from './utils';

/**
 * Request
 * Dispatches different types of request (read/create/delete/update/upload)
 */
class Request {
  constructor(self, schema, site) {
    Object.defineProperty(this, '_self', { writable: false, value: self });
    Object.defineProperty(this, '_schema', { writable: false, value: schema });
    Object.defineProperty(this, '_site', { writable: false, value: site });

    Object.defineProperty(this, '_params', { writable: true, value: {} });

    Object.defineProperty(this, '_track', { writable: true });
    Object.defineProperty(this, '_retry', { writable: true });
    Object.defineProperty(this, '_soft', { writable: true });
    Object.defineProperty(this, '_force', { writable: true });
    Object.defineProperty(this, '_promise', { writable: true });
    Object.defineProperty(this, '_ttl', { writable: true });
    Object.defineProperty(this, '_refresh', { writable: true });
    Object.defineProperty(this, '_expires', { writable: true, value: [] });

    this._dispatched = false;
  }

  /**
   * Get any parameters defined for this request
   * @returns {object} Parameters
   */
  get params() {
    return this._params;
  }

  /**
   * Return the URL for this request
   * @returns {string} URL
   */
  get self() {
    return this._self;
  }

  /**
   * Return the discovered schema for this request
   * @returns {object} Schema
   */
  get schema() {
    return this._schema;
  }

  /**
   * Return the site object that created this request
   * @returns {Site} Site
   */
  get site() {
    return this._site;
  }

  /**
   * Set the tracking key used to monitor this request
   * @param {string|function} key - Tracking key, or function that returns a string to be used as the tracking key
   * @returns {Request} Request
   */
  track(key) {
    this._track = (typeof key === 'function') ? key() : key;
    return this;
  }

  /**
   * Set the maximum amount of time to retry this request for, before rejecting
   * @param {false|number} duration - Retry duration
   * @returns {Request}
   */
  retry(duration) {
    this._retry = duration;
    return this;
  }

  /**
   * Immediately resolve this request if a cached response exists
   * `soft()` or `soft(true)` - return cached response regardless of how old
   * `soft(int)` - return cached response if age not older than the supplied number
   * @param {boolean|number} softness - Softness
   * @returns {Request}
   */
  soft(softness = true) {
    this._soft = softness;
    return this;
  }

  /**
   * Force the request to be sent even if soft() was used
   * This can be used in conjunction with `soft()` to indicate that the content
   * should be updated if possible.
   * @param {boolean} force - Force request dispatch
   * @returns {Request}
   */
  force(force = true) {
    this._force = force;
    return this;
  }

  /**
   * Set whether a Promise should be returned from this request
   * If a promise is not requested and the content does not exist, `null` will
   * be returned.
   * @param {boolean} promise - Return a promise
   * @returns {Request}
   */
  promise(promise = true) {
    this._promise = promise;
    return this;
  }

  /**
   * Set how long this request (and any objects it fetches) should remain in the cache
   * `ttl(false)` - Do not automatically expire the request
   * `ttl(int)` - Automatically expire the request after this amount of time
   * @param {boolean|number} ttl
   * @returns {Request}
   */
  ttl(ttl = false) {
    this._ttl = ttl;
    return this;
  }

  /**
   * Set how often this request should rerun itself automatically
   * @param {boolean|number} refresh
   * @returns {Request}
   */
  refresh(refresh = false) {
    this._refresh = refresh;
    return this;
  }

  /**
   * Mark the result of this request and any fetched objects as expiring when this event occurs
   * @param {string} on - The event to add to the expiry event list
   * @returns {Request}
   */
  expiresOn(on) {
    if (!on || typeof on !== 'string') {
      throw new Error(ERROR_USAGE, 'An event type must be passed to expiresOn');
    }

    this._expires = uniq(append(on, defaultTo([], this._expires)));

    return this;
  }

  /**
   * Remove an event from the list that the request expires on
   * @param {string} off - The event to remove from the expiry event list
   * @returns {Request}
   */
  expiresOff(off) {
    if (!off || typeof off !== 'string') {
      throw new Error(ERROR_USAGE, 'An event type must be passed to expiresOff');
    }

    this._expires = without([off], this._expires);

    return this;
  }

  /**
   * Set a single HTTP request parameter
   * @param {string} key   - Parameter name
   * @param {mixed}  value - Parameter value
   * @returns {Request}
   */
  addParam(key, value) {
    this._params[ key ] = value;

    return this;
  }

  /**
   * Set multiple HTTP request parameters using an object of key/value pairs
   * @param {Object} object - Object of key/value pairs
   * @returns {Request}
   */
  setParams(object) {
    this._params = merge(this._params, object);

    return this;
  }

  /**
   * Remove a single HTTP request parameter
   * @param {string} key - Parameter name
   * @returns {Request}
   */
  removeParam(key) {
    delete this._params[ key ];

    return this;
  }

  /**
   * Dispatch the request and handle the response
   * @param request
   * @returns {*}
   * @private
   */
  _dispatch(request) {
    if (this._dispatched) {
      throw new Error(ERROR_USAGE, 'This request has already been dispatched');
    }

    this._hash = Request._hashRequest(request);

    const action = enqueueRequest(request, {
      track: this._track,
      retry: this._retry,
      soft: this._soft,
      force: this._force,
      promise: this._promise,
      ttl: this._ttl,
      refresh: this._refresh,
      expires: this._expires,
      hash: this._hash,
    });

    this._dispatched = true;

    if (this._promise) {
      return this._site._store.dispatch(action)
        .then((response) => new Response(response, this._site));
    }

    const response = this._site._store.dispatch(action);
    return response ? new Response(response, this._site) : null;
  }

  /**
   * Generate a hash for a request
   * @param {Request} request
   * @returns {*}
   * @private
   */
  static _hashRequest(request) {
    switch (request.method) {
      case METHOD_DOWNLOAD: // Don't include query parameters when generating the hash for a download request
        return objectHash(omit(['query'], request));

      case METHOD_UPLOAD:
        if (isFile(path(['body', 'file'], request))) {
          // Add the blob path to the request so that requests are made unique when using File objects
          request.blob = defaultTo(
            path(['body', 'file', 'name'], request),
            path(['body', 'file', 'preview'], request)
          );

          // Do not use the rest of the file object for hashing
          return objectHash(assocPath(['body', 'file'], null, request));
        }

        return objectHash(request);

      default:
        return objectHash(request);
    }
  }

  /**
   * Dispatch a create request
   * Returns a `Promise` if `promise()` was used, otherwise `null`
   * @param {Object} body - Optional body parameters to pass
   * @returns {Promise|null}
   * @throws Error
   */
  create(body = {}) {
    if (!this.supports(METHOD_CREATE)) {
      throw new Error(ERROR_USAGE, `The endpoint ${this.self} does not support create()`);
    }

    this.setParams(body);

    return this._dispatch({
      url: this.self,
      method: METHOD_CREATE,
      body: this.params,
    });
  }

  /**
   * Dispatch a read request
   * Returns a `Promise` if `promise()` was used, otherwise a `Response` or `null`
   * @param {Object} query - Optional query parameters to pass
   * @returns {Promise|null|Response}
   * @throws Error
   */
  read(query = {}) {
    if (!this.supports(METHOD_READ)) {
      throw new Error(ERROR_USAGE, `The endpoint ${this.self} does not support read()`);
    }

    this.setParams(query);

    return this._dispatch({
      url: this.self,
      method: METHOD_READ,
      query: this.params,
    });
  }

  /**
   * Dispatch an update request
   * Returns a `Promise` if `promise()` was used, otherwise `null`
   * @param {Object} body - Optional body parameters to pass
   * @returns {null|Promise}
   * @throws Error
   */
  update(body = {}) {
    if (!this.supports(METHOD_UPDATE)) {
      throw new Error(ERROR_USAGE, `The endpoint ${this.self} does not support update()`);
    }

    this.setParams(body);

    return this._dispatch({
      url: this.self,
      method: METHOD_UPDATE,
      body: this.params,
    });
  }

  /**
   * Dispatch a delete request
   * Returns a `Promise` if `promise()` was used, otherwise `null`
   * @param {Object} query - Optional query parameters to pass
   * @returns {Promise|null}
   * @throws Error
   */
  delete(query = {}) {
    if (!this.supports(METHOD_DELETE)) {
      throw new Error(ERROR_USAGE, `The endpoint ${this.self} does not support delete()`);
    }

    this.setParams(query);

    return this._dispatch({
      url: this.self,
      method: METHOD_DELETE,
      query: this.params,
    });
  }

  /**
   * Dispatch a file upload request
   * Returns a `Promise` if `promise()` was used, otherwise `null`
   * A `File` object or path must have previously been set with `req.setParams()`
   * @param {Object} body - Optional body parameters to pass
   * @returns {Promise|null}
   * @throws Error
   */
  upload(body = {}) {
    if (!this.supports(METHOD_UPLOAD)) {
      throw new Error(ERROR_USAGE, `The endpoint ${this.self} does not support upload()`);
    }

    this.setParams(body);

    if (!has('file', this.params)) {
      throw new Error(ERROR_USAGE, 'A file must be provided for upload() using set()');
    }

    return this._dispatch({
      url: this.self,
      method: METHOD_UPLOAD,
      body: this.params,
    });
  }

  /**
   * Dispatch a request to download a standard file
   * Returns a `Promise` if `promise()` was used, otherwise `null`
   * Note this may fail if e.g. storage space is exhausted.
   * @returns {Promise|null}
   */
  download() {
    // Check the platform supports download, and that the item is not already downloaded
    if (!platform.supportsDownload()) {
      return maybePromise(false, this._promise);
    }

    return this._dispatch({
      url: this.self,
      method: METHOD_DOWNLOAD,
    });
  }

  /**
   * Cancel a request, so that its promises do not resolve or reject
   */
  cancel() {
    if (!this._promise) {
      warn(REQUEST, 'cancel() was called on a request that did not use a promise');
    }

    if (!this._dispatched) {
      return;
    }

    this._site._store.dispatch(cancelPromise(this._hash));
  }

  /**
   * Test whether this request supports the supplied HTTP method
   * @param {string} method - request technique from const
   * @returns {boolean}
   */
  supports(method) {
    if (this.self === this._site.self && method === METHOD_READ) {
      return true;
    }

    return contains(httpMethod(method), defaultTo([], path(['_schema', 'methods'], this)));
  }
}

/**
 * Default export
 *
 * @private
 */
export default Request;
