import parseLinkHeader from 'parse-link-header';
import { head, isNil, filter, map, mapObjIndexed, omit, path, prop, has } from 'ramda';

import { HTTP_CREATED, HTTP_OK } from './const';
import { hydrateObjects, hydrateOffline } from './selector';
import { getLink, maybeMap } from './utils';

/**
 * Response
 *
 * Wraps response data and provides useful helper functions
 */
class Response {
  constructor(response, site) {
    Object.defineProperty(this, '_site', {
      writable: false,
      value: site,
      enumerable: false,
    });
    Object.defineProperty(this, '_response', {
      writable: false,
      value: response,
      enumerable: false,
    });
    Object.defineProperty(this, '_headers', {
      writable: false,
      value: response.headers,
      enumerable: false,
    });
    Object.defineProperty(this, '_status', {
      writable: false,
      value: response.status,
      enumerable: false,
    });

    // Hydrate the body before trying to use it
    const body = maybeMap((Lbody) => hydrateObjects(Lbody, this._site._store.getState()), response.body);
    Object.defineProperty(this, '_body', {
      writable: false,
      value: body,
      enumerable: false,
    });

    Array.isArray(body)
      ? map((value) => Object.defineProperty(this, value.id, {
        writable: false,
        value,
        enumerable: true,
      }), filter(has('id'), body))
      : mapObjIndexed((value, k) => Object.defineProperty(this, k, {
        writable: false,
        value,
        enumerable: true,
      }), omit(['route', 'link', 'response'], body));
  }

  /**
   * Get the response body
   * @returns {Object}
   */
  get body() {
    return this._body;
  }

  /**
   * Get the response headers
   * @returns {Object}
   */
  get headers() {
    return this._headers;
  }

  /**
   * Get the response status
   * @returns {number}
   */
  get status() {
    return this._status;
  }

  /**
   * Check whether the status is good (not an error code)
   * @returns {boolean}
   */
  get isOk() {
    return this._status === HTTP_OK || this._status === HTTP_CREATED;
  }

  /**
   * Get the total number of pages in this collection
   * @returns {number} Pages
   */
  get totalPages() {
    const totalPages = prop('x-wp-totalpages', this._headers);
    return isNil(totalPages) ? undefined : +totalPages;
  }

  /**
   * Get the total number of results in this collection
   * @returns {number} Results
   */
  get total() {
    const total = prop('x-wp-total', this._headers);
    return isNil(total) ? undefined : +total;
  }

  get linkHeader() {
    const link = prop('link', this._headers);
    return link ? parseLinkHeader(link) : undefined;
  }

  /**
   * Get a referenced link target from this response
   * @param target
   * @returns {string}
   */
  link(target) {
    return getLink(this, target);
  }

  /**
   * Return the first object in this collection (or the target object if not a collection) as a Response object
   * @returns {Object}
   */
  first() {
    return Array.isArray(this._body) ? head(this._body) : this._body;
  }

  /**
   * Returns an endpoint representing the next page in the collection, or undefined if there is no next page
   * @returns {string|undefined}
   */
  next() {
    return path(['next', 'url'], this.linkHeader);
  }

  /**
   * Returns an endpoint representing the previous page in the collection, or undefined if there is no previous page
   * @returns {string|undefined}
   */
  previous() {
    return path(['prev', 'url'], this.linkHeader);
  }

  /**
   * Determine whether this response represents a collection
   * @returns {boolean}
   */
  isCollection() {
    return !!Array.isArray(this._body);
  }

  /**
   * Helper function to return a version of the response body with hrefs replaced with offline versions
   * @returns {Object}
   */
  offline() {
    return hydrateOffline(this._body, this._site._store.getState());
  }
}

export default Response;
