import {
  concat,
  contains,
  defaultTo,
  filter,
  find,
  gt,
  has,
  isNil,
  map,
  path,
  pipe,
  prop,
  propEq,
  reduce,
  reject,
  uniq,
  values,
} from 'ramda';

import { STATUS_UNKNOWN } from './const';
import { rootDir } from './storage';
import { deepValueReplace, isStatusComplete, isURL, stripQuery } from './utils';

/**
 * Get a request using a tracking key
 * @param {string} key   - Tracking key
 * @param {Store}  state - store state
 * @returns {Object}
 */
export const getRequestByTrackingKey = (key, state) =>
  find(propEq('track', key), prop('queue', state));

/**
 * Get expired requests
 * @param {Store} state - store state
 * @returns {array} Array of expired requests
 */
export const getExpiredRequests = (state) =>
  filter((r) => isStatusComplete(r.status)
    && gt(Date.now(), r.ttl ? (r.ttl + r.now) : Date.now()), defaultTo([], prop('queue', state)));

/**
 * Get refreshable requests
 * @param {Store} state - store state
 * @returns {array} Array of refreshable requests
 */
export const getRefreshableRequests = (state) =>
  filter((r) => isStatusComplete(r.status)
    && gt(Date.now(), r.refresh ? (r.refresh + r.now) : Date.now()), defaultTo([], prop('queue', state)));

/**
 * Get expired objects
 * @param {Store} state - store state
 * @returns {*}
 */
export const getExpiredObjects = (state) => {
  const inUse = uniq(reduce((acc, req) => {
    let body = path(['response', 'body'], req);

    if (!body) {
      return acc;
    }

    body = Array.isArray(body) ? body : [body];

    return concat(acc, reject(isNil, map(getSelf, body)));
  }, [], defaultTo([], prop('queue', state))));

  return pipe(
    filter((o) => gt(Date.now(), has('_ttl', o) && o._ttl !== false ? (o._ttl + o._now) : Date.now())),
    filter((o) => !contains(getSelf(o), inUse)),
  )(defaultTo({}, prop('objects', state)));
};

/**
 * Get expired assets
 * @param {Store} state - store state
 * @returns {*}
 */
export const getExpiredAssets = (state) =>
  filter((o) => gt(Date.now(), o.ttl ? (o.ttl + o.now) : Date.now()), defaultTo({}, prop('assets', state)));

/**
 * Get requests in the supplied status
 * @param {string|string[]} statuses - Status or statuses
 * @param {Store}           state    - store state
 * @returns {array} Requests
 */
export const getRequestsByStatus = (statuses, state) =>
  filter((r) => contains(
    r.status,
    (Array.isArray(statuses) ? statuses : [statuses])), defaultTo([], prop('queue', state)),
  );

/**
 * Get the status of a request using a tracking key
 * @param {string} key   - Tracking key
 * @param {Store}  state - store state
 * @returns {string} Request status
 */
export const getRequestStatusByTrackingKey = (key, state) =>
  defaultTo(STATUS_UNKNOWN, prop('status', getRequestByTrackingKey(key, state)));

/**
 * Get an array of requests matching this tracking key
 * @param {string} key   - Tracking key
 * @param {Store}  state - store state
 * @returns {array} Requests
 * @private
 */
export const getRequestsByTrackingKey = (key, state) =>
  filter(propEq('track', key), prop('queue', state));

/**
 * Get a response using a tracking key
 * @param {string} key   - Tracking key
 * @param {Store}  state - store state
 * @returns {object} Response
 */
export const getResponseByTrackingKey = (key, state) =>
  prop('response', defaultTo({}, find(propEq('track', key), prop('queue', state))));

/**
 * Get a request using the parameter hash
 * @param {string} hash  - Parameter hash
 * @param {Store}  state - store state
 * @returns {object} Request
 * @private
 */
export const getRequestByHash = (hash, state) =>
  find(propEq('hash', hash), prop('queue', state));

/**
 * Get a cached object using its self href
 * @param {string} self  - Self href
 * @param {Store}  state - store state
 * @returns {object} Cached object
 */
export const getObjectBySelf = (self, state) =>
  path(['objects', self], state);

/**
 * Get a cached object using a slug and post type
 * @param {string} slug  - slug
 * @param {string} type  - post type
 * @param {Store}  state - store state
 * @returns {Object} Cached object
 */
export const getObjectBySlug = (slug, type, state) => {
  const typeSubstitutions = {
    categories: 'category',
    pages: 'page',
    posts: 'post',
    tags: 'tag',
  };

  type = typeSubstitutions[ type ] || type;

  return find(
    (item) => item.slug === slug && (item.type === type || item.taxonomy === type),
    values(prop('objects', state)),
  );
};

/**
 * Filter to hydrate a response body
 * The system uses a de-duplication method to remove objects from a response
 * body and replace them with references.
 * This selector replaces the references with the original objects.
 * @param {Object} body  - Body
 * @param {Store}  state - store state
 * @returns {Object}
 */
export const hydrateObjects = (body, state) => {
  const self = getSelf(body);

  return !self
    ? body
    : path(['objects', self], state);
};

/**
 * Get the "self" href from one of the known locations
 * @param {Object} body - response body
 * @returns {string|undefined} self href
 */
const getSelf = (body) =>
  defaultTo(path(['_links', 'self', 0, 'href'], body), prop('_self', body));

/**
 * Filter to replace body parameters with offline versions
 * The system can store downloaded items offline.
 * This function replaces all downloaded item references with their offline
 * versions.
 * @param {Object} body  - Body
 * @param {Store}  state - store state
 */
export const hydrateOffline = (body, state) =>
  deepValueReplace(body, map((asset) => `${rootDir()}/assets/${asset.filename}`, state.assets), (value, mapping) => {
    const strippedValue = stripQuery(value);
    return (isURL(value) && strippedValue && has(strippedValue, mapping)) ? prop(strippedValue, mapping) : value;
  });
