import {
  defaultTo,
  endsWith,
  equals,
  has,
  isEmpty,
  keys,
  map,
  omit,
  path,
  prop,
  reduce,
  reduced,
  startsWith,
} from 'ramda';
import URI from 'urijs';
import XRegExp from 'xregexp';
import { event as analyticEvent, pageView as analyticPageView } from './analytics';

import { authenticate, getStatus as getAuthStatus, loggedOut } from './auth';
import {
  check as autoupdateCheck,
  getProgress as getAutoupdateProgress,
  getStatus as getAutoupdateStatus,
} from './autoupdate';
import {
  ANALYTICS,
  ANALYTICS_TYPE,
  AUTH_ME_ENDPOINT,
  AUTH_STATUS_FAILURE,
  AUTH_STATUS_LOGGED_IN,
  AUTH_STATUS_LOGGED_OUT,
  AUTH_STATUS_PENDING,
  AUTH_STATUS_RETRY,
  AUTH_STATUS_UNKNOWN,
  AUTOUPDATE_STATUS_ERROR,
  AUTOUPDATE_STATUS_IDLE,
  AUTOUPDATE_STATUS_PENDING_RESTART,
  AUTOUPDATE_STATUS_UPDATING,
  DAY,
  ERROR_INTERNAL,
  ERROR_USAGE,
  EVENT_AUTH_IDENTITY_CHANGED,
  EVENT_AUTH_LOGGED_IN,
  EVENT_AUTH_LOGGED_OUT,
  EVENT_NOTIFICATION_RECEIVED,
  EVENT_UNHANDLED_REJECTION,
  HOUR,
  HTTP_CREATED,
  HTTP_OK,
  METHOD_READ,
  MINUTE,
  MONTH,
  NETWORK_2G,
  NETWORK_3G,
  NETWORK_4G,
  NETWORK_BLUETOOTH,
  NETWORK_CELLULAR,
  NETWORK_ETHERNET,
  NETWORK_NONE,
  NETWORK_OTHER,
  NETWORK_SLOW_2G,
  NETWORK_UNKNOWN,
  NETWORK_WIFI,
  NETWORK_WIMAX,
  SECOND,
  STATUS_BACKOFF,
  STATUS_FAILED,
  STATUS_INFLIGHT,
  STATUS_READY,
  STATUS_RETRY,
  STATUS_SUCCESS,
  STATUS_UNKNOWN,
  VERSION_APPCONTAINER,
  VERSION_AUTOUPDATE,
  WEEK,
  YEAR,
} from './const';
import { warn } from './debug';
import Error from './error';
import { register as registerEvent, unregister as unregisterEvent } from './event';
import { kvClear, kvRemove, kvSet } from './kv';
import { register as registerNotifications } from './notification';
import { getConnectionType } from './platforms/utils';
import { reboot } from './reboot';
import Request from './request';
import Response from './response';
import {
  getObjectBySelf,
  getObjectBySlug,
  getRequestByHash,
  getRequestByTrackingKey,
  getRequestStatusByTrackingKey,
  getResponseByTrackingKey,
  hydrateObjects,
} from './selector';
import { hide as hideSplashscreen, show as showSplashscreen } from './splashscreen';
import Store from './store';
import { isStatusComplete, isStatusInflight, maybeMap } from './utils';
import { getVersion } from './version';

/**
 * Site
 *
 * @classDesc This is the root service object. An instance of this should be made available throughout the application
 */
class Site {
  /**
   * Create a new Site object
   * The options argument is only used during local development to inform the
   * `Site` object where the service endpoint is.
   * When deployed into a service instance, the endpoint and authentication
   * information are auto-discovered.
   * Note that `start()` should be run on a new instance to fully initialize it.
   * @param {object}  options            - Configuration of this site
   * @param {int}     options.defaultTtl - The default number of milliseconds to store fetched content for
   * @param {int}     options.defaultRetry - The default number of milliseconds to retry fetching content for
   * @param {string}  options.endpoint   - The HTTP endpoint of the instance to connect to
   * @param {string}  options.password   - The password of the developer
   * @param {object}  options.store      - Store object to load at boot
   * @param {string}  options.username   - The username of the developer
   * @param {boolean} options.useMulti   - Whether to use 'multi' requests (default true)
   */
  constructor(options = {}) {
    Object.defineProperty(this, 'self', { writable: true, enumerable: true });
    Object.defineProperty(this, '_options', { value: options });

    // Copy constants to the instance
    map((c) => (this[ c ] = Site[ c ]), [
      'SECOND',
      'MINUTE',
      'HOUR',
      'DAY',
      'WEEK',
      'MONTH',
      'YEAR',
      'VERSION_AUTOUPDATE',
      'VERSION_APPCONTAINER',
      'STATUS_UNKNOWN',
      'STATUS_READY',
      'STATUS_SUCCESS',
      'STATUS_FAILED',
      'STATUS_BACKOFF',
      'STATUS_RETRY',
      'STATUS_INFLIGHT',
      'NOTIFICATION_RECEIVED',
      'UNHANDLED_REJECTION',
      'ERROR_INTERNAL',
      'AUTOUPDATE_STATUS_IDLE',
      'AUTOUPDATE_STATUS_UPDATING',
      'AUTOUPDATE_STATUS_PENDING_RESTART',
      'AUTOUPDATE_STATUS_ERROR',
      'AUTH_STATUS_UNKNOWN',
      'AUTH_STATUS_PENDING',
      'AUTH_STATUS_LOGGED_IN',
      'AUTH_STATUS_LOGGED_OUT',
      'AUTH_STATUS_RETRY',
      'AUTH_STATUS_FAILURE',
      'EVENT_NOTIFICATION_RECEIVED',
      'EVENT_UNHANDLED_REJECTION',
      'EVENT_AUTH_LOGGED_IN',
      'EVENT_AUTH_LOGGED_OUT',
      'EVENT_AUTH_IDENTITY_CHANGED',
      'NETWORK_BLUETOOTH',
      'NETWORK_CELLULAR',
      'NETWORK_ETHERNET',
      'NETWORK_NONE',
      'NETWORK_WIFI',
      'NETWORK_WIMAX',
      'NETWORK_OTHER',
      'NETWORK_UNKNOWN',
      'NETWORK_SLOW_2G',
      'NETWORK_2G',
      'NETWORK_3G',
      'NETWORK_4G',
      'ANALYTICS_TYPE',
    ]);

    this.self = defaultTo(
      prop('endpoint', options),
      prop('__wp_json_endpoint__', typeof window === 'undefined' ? {} : window),
    );

    if (!this.self) {
      throw new Error(ERROR_INTERNAL, 'No endpoint was discovered');
    }

    // Ensure the endpoint is correctly formatted
    if (endsWith('/', this.self)) {
      this.self = this.self.substring(0, this.self.length - 1);
    }

    if (!endsWith('wp-json', this.self)) {
      throw new Error(
        ERROR_USAGE,
        'Endpoint was not specified correctly, should be of the form "https://<hostname>/wp-json"',
      );
    }
  }

  /**
   * Run tasks needed to properly initialize the store
   * Note that in the event of a critical problem (such as network failure
   * during authentication) this function will throw an exception.
   * This event should be handled by the UI.
   * @returns {Promise<*>}
   */
  async start() {
    Object.defineProperty(this, '_store', { value: new Store() });

    await this._store.start({
      site: this,
      endpoint: this.self,
      options: this._options,
    });
  }

  get _routes() {
    return defaultTo({}, path(['schema', 'routes'], this._store.getState()));
  }

  get _namespaces() {
    return defaultTo([], path(['schema', 'namespaces'], this._store.getState()));
  }

  get _schema() {
    return defaultTo({}, prop('schema', this._store.getState()));
  }

  /**
   * Check whether the request being tracked by this key is in flight
   * @param {string} key - Tracking key
   * @returns {boolean}
   */
  isInFlight(key) {
    return isStatusInflight(this.getStatus(key));
  }

  /**
   * Check whether the request being tracked by this key has completed
   * @param {string} key - Tracking key
   * @returns {boolean}
   */
  isComplete(key) {
    return isStatusComplete(this.getStatus(key));
  }

  /**
   * Returns a constant describing the connection type
   * @returns {string} the connection type
   */
  getConnectionType() {
    return getConnectionType();
  }

  /**
   * Return whether the device is online
   * @returns {boolean}
   */
  isOnline() {
    return !equals(getConnectionType(), NETWORK_NONE);
  }

  /**
   * Return the status of this request using its tracking key. Statuses are defined in constants
   * @param {string} key - Tracking key
   * @returns {string}
   */
  getStatus(key) {
    return getRequestStatusByTrackingKey(key, this._store.getState());
  }

  /**
   * Return the current known authentication status according to the store. Statuses are defined in constants
   * @returns {string}
   */
  getAuthStatus() {
    return getAuthStatus(this._store.getState());
  }

  /**
   * Return a plain request using its tracking key
   * @param {string} key - Tracking key
   * @returns {Object}
   */
  getRequest(key) {
    return getRequestByTrackingKey(key, this._store.getState());
  }

  /**
   * Return a plain response using a tracking key
   * @param {string} key - Tracking key
   * @returns {Object|undefined}
   */
  getResponse(key) {
    const response = getResponseByTrackingKey(key, this._store.getState());

    if (!response) {
      return;
    }

    response.body = maybeMap((body) => hydrateObjects(body, this._store.getState()), response.body);

    return response;
  }

  /**
   * Get the body of a response using a tracking key
   * @param {string} key - Tracking key
   * @returns {Object|undefined}
   */
  getResponseBody(key) {
    const response = this.getResponse(key);

    if (!response) {
      return;
    }

    return prop('body', response);
  }

  /**
   * Return a Response object using a tracking key
   * This function is not idempotent
   * @param {string} key - Tracking key
   * @returns {Response|undefined} Response object
   */
  getResponseObject(key) {
    const response = this.getResponse(key);

    if (!response) {
      return;
    }

    return new Response(response, this);
  }

  /**
   * Get a single plain object using a route fragment
   * @param {string} route - Route fragment
   * @returns {Object|undefined}
   */
  getObjectByRoute(route) {
    let request = null;

    try {
      request = this.req(route);
    } catch (err) {
      return;
    }

    if (!request) {
      return;
    }

    return getObjectBySelf(request._self, this._store.getState());
  }

  /**
   * Get a Response object using a route fragment and parameters
   * This function is not idempotent
   * @param {string} route  - Route fragment
   * @param {object} params - Object of key/value pairs
   * @returns {Response|undefined}
   */
  getResponseByRoute(route, params) {
    let request = null;

    try {
      request = this.req(route);
      request.setParams(params);
    } catch (err) {
      return;
    }

    const requestObject = getRequestByHash(
      Request._hashRequest({
        url: request.self,
        method: METHOD_READ,
        query: request.params,
      }),
      this._store.getState(),
    );

    const status = path(['response', 'status'], requestObject);
    if (status !== HTTP_OK && status !== HTTP_CREATED) {
      return;
    }

    return new Response(requestObject.response, this);
  }

  /**
   * Get a collection using a route fragment and parameters
   * @param {string} route  - Route fragment
   * @param {object} params - Object of key/value pairs
   * @returns {Object|undefined}
   */
  getCollectionByRoute(route, params) {
    const response = this.getResponseByRoute(route, params);
    if (response) {
      return response.body;
    }
  }

  /**
   * Get a single plain object using its slug and post type
   * @param {string} slug
   * @param {string} type - post type
   * @returns {Object}
   */
  getObjectBySlug(slug, type) {
    return getObjectBySlug(slug, type, this._store.getState());
  }

  /**
   * Return the supplied plain object as a Response
   * @param object
   * @returns {Response|undefined}
   */
  objectToResponse(object) {
    if (!object) {
      return;
    }

    return new Response({
      body: object,
      status: 200,
      headers: {},
    }, this);
  }

  /**
   * Return the headers from a request object using its tracking key
   * @param {string} key - Tracking key
   * @returns {Object}
   */
  getHeaders(key) {
    const response = getResponseByTrackingKey(key, this._store.getState());

    if (!response) {
      return;
    }

    return prop('headers', response);
  }

  /**
   * Return the schema object
   */
  getSchema() {
    try {
      return this._schema;
    } catch (err) {
      throw new Error(ERROR_USAGE, 'The schema cannot be requested until after start()');
    }
  }

  /**
   * Provides an interface to access the redux store inside this Site
   * @returns {Store}
   */
  get store() {
    return this._store;
  }

  /**
   * Provide an interface to dispatch an action to the store
   * @param {AnyAction} action - Redux Action
   * @returns {*}
   */
  dispatch(action) {
    return this._store.dispatch(action);
  }

  /**
   * Get a section of the key/value store
   * @param {string} key - Key
   * @returns {*} Value
   */
  get(key) {
    return path(['kv', key], this._store.getState());
  }

  /**
   * Set a section of the key/value store
   * @param {Object} object - Key/value object
   * @returns {Site} The Site object for chaining
   */
  set(object) {
    this._store.dispatch(kvSet(object));
    return this;
  }

  /**
   * Remove a key from the key/value store
   * @param {string} key - Key to remove
   * @returns {Site} The Site object for chaining
   */
  remove(key) {
    this._store.dispatch(kvRemove(key));
    return this;
  }

  /**
   * Clear the key/value store
   * @returns {Site} The Site object for chaining
   */
  clear() {
    this._store.dispatch(kvClear());
    return this;
  }

  /**
   * Creates a new request for this Site
   *
   * Request targets can be passed as complete URIs, namespace-prefixed endpoints, or endpoints
   *
   *     s.req('posts')
   *     s.req('posts/1')
   *     s.req('wp/v2/posts/1')
   *     s.req('https://api.example.com/wp-json/wp/v2/posts/1')
   *
   * Request targets can also be passed together with query parameters
   * which will be embedded into the returned Request object
   *
   *     s.req('posts?search=hello')
   *
   * Requests can also be generated from single response objects, in this case checking their embedded 'self' href
   *
   *     s.req(response);
   * @param {string|object} endpoint - The endpoint to run the request against
   * @returns {Request|boolean}
   * @throws {Error}
   */
  req(endpoint) {
    let schema;
    let request = false;

    // noinspection JSUnresolvedFunction
    const searchRoutes = (route) => reduce(
      // eslint-disable-next-line new-cap
      (_, r) => XRegExp(`^${r}$`, 'n')
        .exec(URI(route).search('').toString())
        ? reduced(this._routes[ r ])
        : false,
      false,
      keys(this._routes),
    );

    if (typeof endpoint === 'object') {
      endpoint = path(['_links', 'self', 0, 'href'], endpoint);
    }

    if (!endpoint) {
      throw new Error(ERROR_USAGE, `Invalid route supplied: ${endpoint}`);
    }

    // noinspection JSUnresolvedFunction
    let search = URI(endpoint).search('').toString();

    // Complete URI
    schema = searchRoutes(search.substring(this.self.length));
    if (schema) {
      request = new Request(search, schema, this);
    }

    if (!request && !startsWith('/', search)) {
      search = `/${search}`;
    }

    // Endpoint with namespace (eg /wp/v2/posts/1)
    schema = searchRoutes(search);
    if (schema) {
      request = new Request(this.self + search, schema, this);
    }

    // Endpoint without namespace (eg /posts/1)
    schema = reduce((acc, namespace) => {
      const Lschema = searchRoutes(`/${namespace}${search}`);
      return Lschema ? reduced({ namespace, schema: Lschema }) : false;
    }, false, this._namespaces);
    if (schema) {
      request = new Request(`${this.self}/${schema.namespace}${search}`, schema.schema, this);
    }

    // Schemaless request
    if (!request) {
      request = new Request(endpoint, {}, this);
    }

    // noinspection JSUnresolvedFunction
    request.setParams(URI(endpoint).search(true));

    return request;
  }

  /**
   * Reboots the application
   * @returns {void}
   */
  reboot() {
    this._store.dispatch(reboot());
  }

  /**
   * Perform manual authentication
   * Authentication can be performed with any of: {username,password}, {email,password}, {jwt}
   * @param {Object} args - authentication parameters
   * @returns {Promise}
   */
  login(args = {}) {
    return this._store.dispatch(authenticate(args));
  }

  /**
   * Perform logout
   * @returns {Promise}
   */
  logout() {
    return this._store.dispatch(loggedOut());
  }

  /**
   * Determine whether the user is logged in
   *
   * Returns
   *  - `true` if the user is known to be logged in
   *  - `false` if they are known to be logged out
   *  - `undefined` if the login state is not known (eg when offline)
   * @param {boolean} soft - Send a network request to confirm whether the user is logged in
   * @returns {boolean|undefined|Promise}
   */
  isLoggedIn(soft = true) {
    if (soft) {
      return getAuthStatus(this._store.getState()) === AUTH_STATUS_LOGGED_IN;
    }

    return this.req(AUTH_ME_ENDPOINT)
      .promise()
      .read()
      .then(() => true)
      .catch(() => false);
  }

  /**
   * Register for push notifications
   * @returns {void}
   */
  registerNotifications() {
    this._store.dispatch(registerNotifications());
  }

  /**
   * Register an event handler
   * @param event
   * @param callback
   * @returns Site
   */
  on(event, callback) {
    if (typeof callback !== 'function') {
      throw new Error(ERROR_USAGE, 'Event registration was passed a callback that was not a function');
    }

    if (typeof event !== 'string') {
      throw new Error(ERROR_USAGE, 'Event registration was passed an invalid event type');
    }

    this._store.dispatch(registerEvent(event, callback));
    return this;
  }

  /**
   * Register an event handler that should run only once
   * @param event
   * @param callback
   * @returns Site
   */
  one(event, callback) {
    if (typeof callback !== 'function') {
      throw new Error(ERROR_USAGE, 'Event registration was passed a callback that was not a function');
    }

    this._store.dispatch(registerEvent(event, callback, true));
    return this;
  }

  /**
   * Remove an event handler
   * @param event
   * @param callback
   * @returns Site
   */
  off(event, callback) {
    this._store.dispatch(unregisterEvent(event, callback));
    return this;
  }

  /**
   * Get a printable version, optionally of a specific component
   * @param component
   * @returns {*}
   */
  getVersion(component) {
    return getVersion(component, this._store.getState());
  }

  /**
   * Show splash screen
   */
  showSplashscreen() {
    this._store.dispatch(showSplashscreen());
  }

  /**
   * Hide splash screen
   */
  hideSplashscreen() {
    this._store.dispatch(hideSplashscreen());
  }

  /**
   * Get the status of the autoupdate service
   * @returns {string}
   */
  getUpdateStatus() {
    return getAutoupdateStatus(this._store.getState());
  }

  /*
   * Force an update request
   */
  autoupdate() {
    this._store.dispatch(autoupdateCheck(true));
  }

  /**
   * Get the progress of the autoupdate service
   * @returns {number}
   */
  getUpdateProgress() {
    return getAutoupdateProgress(this._store.getState());
  }

  /**
   * Add a store subscription function
   * @param {function} f
   * @returns {Site}
   */
  sub(f) {
    this._store.subscribe(f);
    return this;
  }

  /**
   * Track a page view / navigation change in analytics services
   * ```
   * trackPageView('/home');
   * trackPageView('/home', { title: 'Home Page' });
   * trackPageView('/articles/22/new-product-launch', { objectRoute: 'article/22' });
   * ```
   * @param {string} resource - The path or resource of the page being viewed
   * (e.g. `/home`).
   * @param {object} [args] - Extra arguments for the page view.
   * @param {string} [args.title] - The title of the view being accessed (e.g.
   * "Home" or "Preferences"). If no title is provided then one will be derived
   * from `objectRoute` (if provided).
   * @param {string} [args.objectRoute] - The API route of the object being
   * viewed (e.g. `article/5`).
   * @param {object} [args.google] - Extra properties passed to Google Analytics
   * @returns {undefined}
   */
  trackPageView(resource, args = {}) {
    if (typeof resource === 'undefined') {
      warn(ANALYTICS, 'Submitted page views must have a resource');
      return;
    }

    if (!isEmpty(omit(['title', 'objectRoute', 'google'], args))) {
      warn(ANALYTICS, 'The submitted page view contained unknown props');
      return;
    }

    this._store.dispatch(analyticPageView(resource, args));
  }

  /**
   * Track an event for analytics services
   * ```
   * trackEvent(Site.ANALYTICS_TYPE.SOCIAL, { action: 'rate', value: 0.4, objectRoute: 'article/4' });
   * trackEvent(Site.ANALYTICS_TYPE.GENERIC, { action: 'submit', label: 'Submit' });
   * trackEvent(Site.ANALYTICS_TYPE.SEARCH, { action: 'query', label: 'new product' });
   * trackEvent(Site.ANALYTICS_TYPE.MEDIA, { action: 'play', value: 20, objectRoute: 'media/234' });
   * trackEvent(Site.ANALYTICS_TYPE.MEDIA, { action: 'pause', objectRoute: 'media/234' });
   * ```
   * @param {number} type - Event category (e.g. social, app, notification).
   * Should be a member of the `ANALYTICS_TYPE` enum.
   * @param {object} [event] - The event object.
   * @param {string} [event.action] - The action that caused the event typically
   * as a verb (e.g. "search", "like", "play").
   * @param {number} [event.value] - A value of action (e.g., a rating value,
   * search terms, or play head location) as a non-negative integer.
   * @param {string} [event.label] - A label for the action.
   * @param {string} [event.objectRoute] - The API route of the object being
   * operated on (e.g. `article/4` or `media/5`).
   * @param {boolean} [event.interactive = true] - Whether this event was caused
   * by direct user interaction.
   * @param {object} [event.google] - Extra properties passed to Google
   * Analytics.
   * @returns {undefined}
   */
  trackEvent(type, event = {}) {
    if (typeof type === 'undefined') {
      warn(ANALYTICS, 'Submitted events must have a type');
      return;
    }

    if (typeof type !== 'number') {
      warn(ANALYTICS, 'Submitted event did not have a numeric type');
      return;
    }

    if (
      event.value
      && (
        typeof event.value !== 'number'
        || event.value < 0
        || event.value !== parseInt(event.value, 10)
      )
    ) {
      warn(ANALYTICS, 'Submitted event value must be a non-negative integer');
      return;
    }

    for (const arg of ['action', 'label', 'objectRoute']) {
      if (has(arg, event) && typeof event[ arg ] !== 'string') {
        warn(ANALYTICS, `The submitted event argument "${arg}" was not provided as a string`);
        return;
      }
    }

    this._store.dispatch(analyticEvent(type, event));
  }

  /**
   * Sanitize an input string to be used as a slug
   * @param {string} s
   * @returns {string}
   */
  toSlug(s) {
    return s.toLowerCase()
      .replace(/[^%a-z0-9 _-]/g, '')
      .replace(/\s+/g, '-')
      .replace(/-+/g, '-')
      .replace(/^-+|-+$/g, '');
  }
}

Site.SECOND = SECOND;
Site.MINUTE = MINUTE;
Site.HOUR = HOUR;
Site.DAY = DAY;
Site.WEEK = WEEK;
Site.MONTH = MONTH;
Site.YEAR = YEAR;

Site.VERSION_APP = VERSION_AUTOUPDATE;
Site.VERSION_CONTAINER = VERSION_APPCONTAINER;

Site.STATUS_UNKNOWN = STATUS_UNKNOWN;
Site.STATUS_READY = STATUS_READY;
Site.STATUS_SUCCESS = STATUS_SUCCESS;
Site.STATUS_FAILED = STATUS_FAILED;
Site.STATUS_BACKOFF = STATUS_BACKOFF;
Site.STATUS_RETRY = STATUS_RETRY;
Site.STATUS_INFLIGHT = STATUS_INFLIGHT;

Site.ERROR_INTERNAL = ERROR_INTERNAL;

Site.AUTH_STATUS_UNKNOWN = AUTH_STATUS_UNKNOWN;
Site.AUTH_STATUS_PENDING = AUTH_STATUS_PENDING;
Site.AUTH_STATUS_LOGGED_IN = AUTH_STATUS_LOGGED_IN;
Site.AUTH_STATUS_LOGGED_OUT = AUTH_STATUS_LOGGED_OUT;
Site.AUTH_STATUS_RETRY = AUTH_STATUS_RETRY;
Site.AUTH_STATUS_FAILURE = AUTH_STATUS_FAILURE;

Site.AUTOUPDATE_STATUS_IDLE = AUTOUPDATE_STATUS_IDLE;
Site.AUTOUPDATE_STATUS_UPDATING = AUTOUPDATE_STATUS_UPDATING;
Site.AUTOUPDATE_STATUS_PENDING_RESTART = AUTOUPDATE_STATUS_PENDING_RESTART;
Site.AUTOUPDATE_STATUS_ERROR = AUTOUPDATE_STATUS_ERROR;

Site.EVENT_NOTIFICATION_RECEIVED = EVENT_NOTIFICATION_RECEIVED;
Site.EVENT_UNHANDLED_REJECTION = EVENT_UNHANDLED_REJECTION;
Site.EVENT_AUTH_LOGGED_IN = EVENT_AUTH_LOGGED_IN;
Site.EVENT_AUTH_LOGGED_OUT = EVENT_AUTH_LOGGED_OUT;
Site.EVENT_AUTH_IDENTITY_CHANGED = EVENT_AUTH_IDENTITY_CHANGED;

Site.NETWORK_BLUETOOTH = NETWORK_BLUETOOTH;
Site.NETWORK_CELLULAR = NETWORK_CELLULAR;
Site.NETWORK_ETHERNET = NETWORK_ETHERNET;
Site.NETWORK_NONE = NETWORK_NONE;
Site.NETWORK_WIFI = NETWORK_WIFI;
Site.NETWORK_WIMAX = NETWORK_WIMAX;
Site.NETWORK_OTHER = NETWORK_OTHER;
Site.NETWORK_UNKNOWN = NETWORK_UNKNOWN;
Site.NETWORK_SLOW_2G = NETWORK_SLOW_2G;
Site.NETWORK_2G = NETWORK_2G;
Site.NETWORK_3G = NETWORK_3G;
Site.NETWORK_4G = NETWORK_4G;

Site.ANALYTICS_TYPE = ANALYTICS_TYPE;

/**
 * Default export
 *
 * @private
 */
export default Site;
