import { assoc, evolve, map, propEq, when } from 'ramda';

import { STATUS_INFLIGHT, STATUS_RETRY } from './const';
import * as platform from './platforms/storage';

/**
 * Caches the JSON-encoded store so we can determine when it has changed
 * @type {string}
 */
let lastDehydratedState = '';

/**
 * Loads the store from long-term storage
 * @returns {Promise<any>} parsed store data
 */
export const hydrate = async () => {
  const store = await platform.load();

  try {
    return JSON.parse(store);
  } catch (err) {
    return null;
  }
};

/**
 * Writes the store to long-term storage
 * @param {Store} state - Redux store
 * @returns {*}
 */
export const dehydrate = async (state) => {
  // Move all inflight requests to the retry state
  state = evolve({
    queue: map(when(
      propEq('status', STATUS_INFLIGHT),
      assoc('status', STATUS_RETRY),
    )),
  }, state);

  state = JSON.stringify(state, null, 2);

  // If the store has not changed, return
  if (state === lastDehydratedState) {
    return;
  }

  // Update with the latest store content
  lastDehydratedState = state;

  // Write the last known store
  platform.save(state);
};

export {
  rootDir,
  readFile,
  writeFile,
  fileExists,
  directoryExists,
  moveDirectory,
  deleteDirectory,
  recursiveCopy,
  deleteFile,
} from './platforms/storage';
