import { compose, defaultTo, isEmpty, pick, prop } from 'ramda';
import { createStore } from 'redux';
import { init as initAnalytics } from './analytics';

import { setBasicCredentials, setTokenCredentials } from './auth';
import { init as autoUpdateInit } from './autoupdate';
import { refresh as refreshConnection } from './connection';
import { AUTH_ME_ENDPOINT, EVENT_UNHANDLED_REJECTION, HOUR, NOOP, SECOND, STORE } from './const';
import { error, info } from './debug';
import { generateEvent } from './event';
import { set as setMeta } from './meta';
import { middleware } from './middleware';
import { refresh as refreshNotifications } from './notification';
import * as platform from './platforms/store';
import { Promise } from './promise';
import { queueProcess } from './queue';
import { reducers, serialize } from './reducers';
import { fetch as fetchSchema } from './schema';
import { getRefreshableRequests } from './selector';
import { hide as hideSplashScreen } from './splashscreen';
import { dehydrate, hydrate } from './storage';
import { externalFlag } from './utils';
import { refresh as refreshVersion } from './version';

class Store {
  async start(meta = {}) {
    const { site, endpoint, options = {} } = meta;

    Promise.onPossiblyUnhandledRejection((err) => this.store.dispatch(generateEvent(EVENT_UNHANDLED_REJECTION, err)));

    Promise.config({
      warnings: false,
    });

    const enhancers = [middleware, ...platform.getEnhancers()];

    await platform.ready();

    let hydrationTime = Date.now();

    const initialStore = defaultTo(await hydrate(), options.store);

    // noinspection JSValidateTypes
    this.store = createStore(
      reducers,
      defaultTo(reducers({}, { type: NOOP }), initialStore),
      compose(...enhancers),
    );
    hydrationTime = Date.now() - hydrationTime;

    // Subscribe to store updates to serialize
    if (!options.store) {
      this.store.subscribe(() => {
        // noinspection JSIgnoredPromiseFromCall
        dehydrate(pick(serialize, this.store.getState()));
      });
    }

    // Store site
    this.store.dispatch(setMeta({ site, endpoint }));

    // Setup optional authentication
    const { username, password } = options;
    if (username && password) {
      this.store.dispatch(setBasicCredentials(username, password));
    }

    if (options.authtoken) {
      this.store.dispatch(setTokenCredentials(options.authtoken));
    }

    if (externalFlag('debug')) {
      const storeSize = Math.round(JSON.stringify(this.store.getState()).length / 1024);
      info(STORE, `Store ready after ${hydrationTime}ms, ${storeSize}KB stored`);
    }

    // Ensure a schema exists
    if (isEmpty(prop('schema', this.store.getState()))) {
      await this.store.dispatch(fetchSchema());
    } else {
      this.store.dispatch(fetchSchema())
        .catch((e) => {
          if (!externalFlag('simulateOffline')) {
            error(STORE, 'Failed to fetch schema', e);
          }
        });
    }

    // Setup autoupdate
    await this.store.dispatch(autoUpdateInit(options.repository));

    // Periodically check authentication status
    await site.req(AUTH_ME_ENDPOINT)
      .soft()
      .force()
      .ttl(false)
      .refresh(HOUR)
      .promise()
      .read()
      .catch(() => {});

    // Update notifications
    this.store.dispatch(refreshNotifications());

    // Update connection status
    this.store.dispatch(refreshConnection());

    // Update versions
    this.store.dispatch(refreshVersion());

    // Setup analytics
    this.store.dispatch(initAnalytics());

    // Add event handlers
    platform.addEventListeners(this.store.dispatch);

    // Check the queue for refresh requests
    if (platform.supportsScheduledTasks()) {
      setInterval(() => {
        if (!isEmpty(getRefreshableRequests(this.store.getState()))) {
          this.store.dispatch(queueProcess());
        }
      }, SECOND);
    }

    // Hide the splash screen
    this.store.dispatch(hideSplashScreen());
  }

  /**
   * Dispatches an action
   * @param {AnyAction} action - Redux action
   * @returns {AnyAction} For convenience, the same action object you dispatched
   */
  dispatch(action) {
    return this.store.dispatch(action);
  }

  /**
   * Reads the state tree managed by the store
   * @returns {Store} The current state tree of your application
   */
  getState() {
    return this.store.getState();
  }

  /**
   * Adds a change listener
   * @param {function(): void} listener - A callback to be invoked on every dispatch
   * @returns {Unsubscribe} A function to remove this change listener
   */
  subscribe(listener) {
    return this.store.subscribe(listener);
  }
}

export default Store;
