import {
  assoc,
  contains,
  curry,
  defaultTo,
  filter,
  fromPairs,
  has,
  is,
  keys,
  map,
  path,
  pipe,
  prop,
  reduce,
  replace,
  split,
  startsWith,
} from 'ramda';
import URI from 'urijs';
import uuid from 'uuid';

import {
  BACKOFF_BASE,
  BACKOFF_CAP,
  METHOD_CREATE,
  METHOD_DELETE,
  METHOD_DOWNLOAD,
  METHOD_MULTI,
  METHOD_READ,
  METHOD_UPDATE,
  METHOD_UPLOAD,
  PLATFORM_BROWSER,
  PLATFORM_CORDOVA,
  PLATFORM_NODEJS,
  SECOND,
  STATUS_BACKOFF,
  STATUS_FAILED,
  STATUS_INFLIGHT,
  STATUS_READY,
  STATUS_RETRY,
  STATUS_SUCCESS,
} from './const';
import * as platformSupport from './platforms/utils';
import { Promise } from './promise';

export const backoff = (retries) =>
  Math.random() * (Math.min(BACKOFF_CAP, BACKOFF_BASE * Math.pow(2, defaultTo(0, retries))) + 1) * SECOND;

export const maybeMap = (f, v) => Array.isArray(v) ? map(f, v) : f(v);

export const maybePromise = (result, promisify) => promisify ? Promise.resolve(result) : result;

/**
 * Determines the platform on which we're running
 * @returns {PLATFORM_BROWSER | PLATFORM_CORDOVA | PLATFORM_NODEJS}
 */
export const platform = () =>
  typeof window === 'undefined' ? PLATFORM_NODEJS : typeof window.cordova !== 'undefined'
    ? PLATFORM_CORDOVA
    : PLATFORM_BROWSER;

/**
 * Converts a DataLayer HTTP method constant into the name of the method
 * @param {HttpMethod} method - HTTP method constant
 * @returns {string}
 */
export const httpMethod = (method) => {
  switch (method) {
    case METHOD_DOWNLOAD:
    case METHOD_READ:
      return 'GET';

    case METHOD_UPDATE:
      return 'PUT';

    case METHOD_CREATE:
    case METHOD_MULTI:
    case METHOD_UPLOAD:
      return 'POST';

    case METHOD_DELETE:
      return 'DELETE';
  }
};

export const getLink = (response, target) => path(['_links', target, 0, 'href'], response);

/**
 * Function that takes a search object, a mapping of keys to values and optionally a replacer function
 * Iterates over the entire object running the mapping function on each value.
 * Default replacer function looks for exact match between the object value and
 * the mapping key and replaces the object value with the mapped value.
 * @param search
 * @param mapping
 * @param replacer
 * @returns {Function}
 * @private
 */
export const deepValueReplace = (
  search, mapping, replacer = ((v, Lmapping) => has(v, Lmapping) ? prop(v, Lmapping) : v),
) => map((v) => is(Object, v) ? deepValueReplace(v, mapping, replacer) : replacer(v, mapping), search);

// noinspection JSUnresolvedFunction
export const stripQuery = (href) =>
  isURL(href) ? URI(href).search('').toString() : href;

/**
 * Checks whether a supplied object looks like a URL
 * @param {any} href - potential URL to test
 * @returns {boolean}
 */
export const isURL = (href) => startsWith('http', `${href}`);

export const isStatusComplete = (status) => contains(status, [STATUS_FAILED, STATUS_SUCCESS]);

export const isStatusInflight = (status) =>
  contains(status, [STATUS_READY, STATUS_BACKOFF, STATUS_RETRY, STATUS_INFLIGHT]);

/**
 * Encodes a string using the Base64 encoding scheme
 * @param {string} string - string to encode
 * @returns {string} Base64-encoded version of input string
 */
export { base64encode } from './platforms/utils';

export const debounce = (func, wait, immediate) => {
  let timeout;

  return function (...args) {
    const context = this;

    const later = function () {
      timeout = null;
      if (!immediate) func.apply(context, args);
    };

    const callNow = immediate && !timeout;
    clearTimeout(timeout);
    timeout = setTimeout(later, wait);

    if (callNow) func.apply(context, args);
  };
};

export const isFile = (file) =>
  file
  && (
    (typeof File !== 'undefined' && file instanceof File)
    || (typeof file === 'object' && file.constructor.name === 'File')
  );

/**
 * Retrieves a configuration flag
 * Flags are read from the `storyshareConfig` string in `localStorage` or the
 * `STORYSHARE_CONFIG` environment variable depending on platform.
 * Some flags are boolean, and will be `true` if present (e.g. `flag`).
 * Some flags take values specified using `=` (e.g. `flag = value`).
 * Any flags not present are considered to be `false`.
 * Multiple flags can be separated using commas, and spaces are ignored.
 * @param {string} flag - Name of flag to retrieve
 * @returns {boolean|string} flag value
 */
export const externalFlag = (flag) => {
  const flags = platformSupport.readFlags() || '';
  const flagsObject = pipe(
    replace(/ /g, ''),
    split(','),
    map(split('=')),
    fromPairs(),
    map((value) => value || true), // flags default to true if present
  )(flags);

  return flagsObject[ flag ] || false; // flags default to false if not found
};

/**
 * Return the current viewport size
 * @returns {null|object}
 */
export const getViewportSize = () => {
  if (typeof document === 'undefined' || typeof window === 'undefined') {
    return null;
  }

  return {
    width: Math.max(document.documentElement.clientWidth, window.innerWidth || 0),
    height: Math.max(document.documentElement.clientHeight, window.innerHeight || 0),
  };
};

/**
 * Creates a new object with the own properties of the provided object, but the
 * keys renamed according to the keysMap object as `{oldKey: newKey}`.
 * When some key is not found in the keysMap, then it's passed as-is.
 *
 * https://github.com/ramda/ramda/wiki/Cookbook#rename-keys-of-an-object
 *
 * Keep in mind that in the case of keys conflict is behaviour undefined and
 * the result may vary between various JS engines!
 *
 * @sig {a: b} -> {a: *} -> {b: *}
 */
export const renameKeys = curry((keysMap, obj) =>
  reduce((acc, key) => assoc(keysMap[ key ] || key, obj[ key ], acc), {}, keys(obj)),
);

/**
 * Strip undefined properties from the provided object
 * @param {object} obj - The object that will have undefined props removed
 * @returns {object} - A copy of the object with undefined props removed
 */
export const stripUndefinedProps = (obj) => filter((value) => typeof value !== 'undefined', obj);

/**
 * Return a UUID
 * @returns {string}
 */
export const generateHexUUID = () => uuid.v4().replace(/-/g, '');
